import os
import pickle
from attrs import define, field
from datetime import datetime
from typing import Optional, Tuple, Union, Dict, List

import numpy as np
import pandas as pd
import torch

from botorch.test_functions.synthetic import (
    SyntheticTestFunction,
    Ackley,
    Griewank,
    Levy,
    Michalewicz,
    Rastrigin,
    Rosenbrock,
    StyblinskiTang,
)

from gpytorch.kernels import AdditiveStructureKernel, Kernel, MaternKernel, ScaleKernel

from pymoo.core.problem import Problem
from pymoo.factory import get_reference_directions, get_termination
from pymoo.optimize import minimize
from pymoo.problems.multi import Kursawe, ZDT1, ZDT2, ZDT3, ZDT4, ZDT6

from nemo_bo.acquisition_functions.nsga_improvement.unsga3 import UNSGA3
import nemo_bo.utils.logger as logging_nemo
from nemo_bo.opt.objectives import CalculableObjective, ObjectivesList, RegressionObjective
from nemo_bo.opt.variables import VariablesList
from nemo_bo.opt.constraints import ConstraintsList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define
class Benchmark:
    """

    Benchmark functions are typically used to simulate the outcomes of experiments in a closed-loop manner and therefore
    can be helpful to evaluate the quality of an optimisation (inferred from the effectiveness of the utilised ML
    model(s) and/or acquisition function to identify the optimum)

    """

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Returns the corresponding output values and standard deviations using the inputted X array and benchmark method

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation

        """
        pass

    def best_result(self) -> Optional[Union[int, float, np.ndarray]]:
        """

        Returns the best result for the benchmark function

        """
        pass


@define
class SingleObjectiveSyntheticBenchmark(Benchmark):
    """

    Adopts the following synthetic functions from the BoTorch package to use as functions in a benchmark.
    M. Balandat, B. Karrer, D. R. Jiang, S. Daulton, B. Letham, A. G. Wilson, and E. Bakshy. BoTorch: A Framework for
    Efficient Monte-Carlo Bayesian Optimization. Advances in Neural Information Processing Systems 33, 2020.
    http://arxiv.org/abs/1910.06403
    https://github.com/pytorch/botorch
    https://botorch.org/

    All functions are to be minimized. The recommended bounds for each variable is shown below:

    1. Ackley (var_bounds = [-32.768, 32.768], obj_bounds = [0.0, 25.0])
    2. Griewank (var_bounds = [-600.0, 600.0], obj_bounds = [0.0, 100.0 * n_var])
    3. Levy (var_bounds = [-10.0, 10.0], obj_bounds = [0.0, 50.0 * n_var])
    4. Michalewicz (var_bounds = [0.0, math.pi], n_var = 2, 5, or 10, obj_bounds = [-1.9, 0.0], [-4.8, 0.0], [-9.8, 0.0])
    5. Rastrigin (var_bounds = [-5.12, 5.12], obj_bounds = [0.0, 50.0 * n_var])
    6. Rosenbrock (var_bounds = [-5.0, 10.0], obj_bounds = [0.0, 810081.0 * (n_var - 1)])
    7. StyblinskiTang (var_bounds = [-5.0, 5.0], obj_bounds = [-40.0 * n_var, 25.0])

    Parameters
    ----------
    function_name: str
        The name of the synthetic function to use ('ackley', 'griewank', 'levy', 'michalewicz',
        'rastrigin', 'rosenbrock', or 'styblinski-tang')
    dim: int
        The number of input dimensions
    noise_std: float = None
        Standard deviation of the observation noise

    """

    function_name: str
    dim: int
    noise_std: Optional[float] = None
    fitted: bool = field(init=False)
    synthetic_function: SyntheticTestFunction = field(init=False)

    def __attrs_post_init__(self):
        if self.function_name == "ackley":
            self.synthetic_function = Ackley(dim=self.dim, noise_std=self.noise_std)
        elif self.function_name == "griewank":
            self.synthetic_function = Griewank(dim=self.dim, noise_std=self.noise_std)
        elif self.function_name == "levy":
            self.synthetic_function = Levy(dim=self.dim, noise_std=self.noise_std)
        elif self.function_name == "michalewicz":
            if self.dim not in [2, 5, 10]:
                raise ValueError(
                    f"The input of input dimensions has been set as {self.dim}. Please use 2, 5, or 10 "
                    f"input dimensions for the Michalewicz function"
                )
            self.synthetic_function = Michalewicz(dim=self.dim, noise_std=self.noise_std)
        elif self.function_name == "rastrigin":
            self.synthetic_function = Rastrigin(dim=self.dim, noise_std=self.noise_std)
        elif self.function_name == "rosenbrock":
            self.synthetic_function = Rosenbrock(dim=self.dim, noise_std=self.noise_std)
        elif self.function_name == "styblinski-tang":
            self.synthetic_function = StyblinskiTang(dim=self.dim, noise_std=self.noise_std)

        self.fitted = True

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Calculate the corresponding output values using the desired synthetic function and the inputted X array

        Parameters
        ----------
        X: np.ndarray
            X array that contains only numerical data

        """
        Y = self.synthetic_function.evaluate_true(torch.tensor(X)).cpu().detach().numpy()

        return Y, np.zeros_like(Y)

    def best_result(self) -> float:
        """

        Returns the best result for the benchmark function

        """
        if self.function_name == "ackley":
            return 0.0
        elif self.function_name == "griewank":
            return 0.0
        elif self.function_name == "levy":
            return 0.0
        elif self.function_name == "michalewicz":
            if self.dim == 2:
                return -1.8013
            elif self.dim == 5:
                return -4.687658
            elif self.dim == 10:
                return -9.66015
        elif self.function_name == "rastrigin":
            return 0.0
        elif self.function_name == "rosenbrock":
            return 0.0
        elif self.function_name == "styblinski-tang":
            return -39.16599 * self.dim


@define
class MultiObjectiveSyntheticBenchmark(Benchmark):
    """

    Adopts the following synthetic functions from the Pymoo package to use as functions in a benchmark.
    The NSGA-based methods were copied and adapted from J. Blank and K. Deb, pymoo: Multi-Objective Optimization in
    Python, in IEEE Access, vol. 8, pp. 89497-89509, 2020, doi: 10.1109/ACCESS.2020.2990567
    https://ieeexplore.ieee.org/document/9078759
    https://github.com/anyoptimization/pymoo

    All objectives in every function are to be minimized. The recommended bounds for each dimension, the number of
    variables, and the number of objectives is shown in parentheses:

    1. Kursawe (var_bounds = [-5.0, 5.0], obj_bounds = approx f1: [-20.0, -5.0], f2: [-13.0, 23.0], n_var = 3,
    n_obj = 2)
    2. ZDT1 (var_bounds = [0.0, 1.0], obj_bounds = approx f1: [0.0, 1.0], f2: [0.0, 7.5], n_var = 30, n_obj = 2)
    3. ZDT2 (var_bounds = [0.0, 1.0], obj_bounds = approx f1: [0.0, 1.0], f2: [0.0, 8.0], n_var = 30, n_obj = 2)
    4. ZDT3 (var_bounds = [0.0, 1.0], obj_bounds = approx f1: [0.0, 1.0], f2: [0.0, 7.5], n_var = 30, n_obj = 2)
    5. ZDT4 (var_bounds = var1: [0.0, 1.0], var2 to var10: [-10.0, 10.0], obj_bounds = approx f1: [0.0, 1.0], f2: [0.0, 900.0], n_var = 10, n_obj = 2)
    6. ZDT6 (var_bounds = [0.0, 1.0], obj_bounds = approx f1: [0.0, 1.0], f2: [0.0, 10.0], n_var = 10, n_obj = 2)

    Parameters
    ----------
    function_name: str
        The name of the synthetic function to use ('kursawe', 'zdt1', 'zdt2', 'zdt3', 'zdt4', 'zdt6')

    """

    function_name: str
    fitted: bool = field(init=False)
    synthetic_function: Problem = field(init=False)

    def __attrs_post_init__(self):
        if self.function_name == "kursawe":
            self.synthetic_function = Kursawe()
        elif self.function_name == "zdt1":
            self.synthetic_function = ZDT1()
        elif self.function_name == "zdt2":
            self.synthetic_function = ZDT2()
        elif self.function_name == "zdt3":
            self.synthetic_function = ZDT3()
        elif self.function_name == "zdt4":
            self.synthetic_function = ZDT4()
        elif self.function_name == "zdt6":
            self.synthetic_function = ZDT6()
        else:
            raise ValueError(f"The synthetic function called {self.function_name} is not supported")

        self.fitted = True

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Calculate the corresponding output values using the desired synthetic function and the inputted X array

        Parameters
        ----------
        X: np.ndarray
            X array that contains only numerical data

        """
        if self.function_name == "kursawe":
            if X.shape[1] != 3:
                raise AttributeError(
                    f"The X array contains {X.shape[1]} variables but the Kursawe two objective "
                    f"synthetic function is for 3 variables"
                )
        elif self.function_name in ["zdt1", "zdt2", "zdt3"]:
            if X.shape[1] != 30:
                raise AttributeError(
                    f"The X array contains {X.shape[1]} variables but the ZDT1, ZDT2, and ZDT3 two objective "
                    f"synthetic functions are for 30 variables"
                )
        elif self.function_name in ["zdt4", "zdt6"]:
            if X.shape[1] != 10:
                raise AttributeError(
                    f"The X array contains {X.shape[1]} variables but the ZDT4, and ZDT6 two objective "
                    f"synthetic functions are for 10 variables"
                )

        out = {"F": np.array([0])}  # Dummy array as the value for 'F'
        self.synthetic_function._evaluate(X, out)
        Y = out["F"]

        return Y, np.zeros_like(Y)

    def best_result(self) -> np.ndarray:
        """

        Returns the best result for the benchmark function

        """
        return self.synthetic_function._calc_pareto_front()


@define
class ModelBenchmark(Benchmark):
    """

    This type of benchmark object is used to accurately predict the outcome of optimisation suggestions as they
    typically utilise all available data for regression objectives. This method will also work for the calculable
    objectives in the optimisation problem.

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    objectives: ObjectivesList
        ObjectivesList object that contains information about all objectives
    constraints: ConstraintsList, Default = None
        ConstraintsList object that contains information about all constraints

    """

    variables: VariablesList
    objectives: ObjectivesList
    constraints: Optional[ConstraintsList] = None
    fitted: bool = field(init=False)
    best: np.ndarray = field(init=False)
    X: np.ndarray = field(init=False)
    Y: np.ndarray = field(init=False)

    def __attrs_post_init__(self):
        self.fitted = False

    def fit(self, X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, **kwargs) -> None:
        """

        Fits ML models for all specified regression objectives to be used a benchmark functions

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        test_ratio: float
            The proportion of inputted X and Y arrays to be split for the validation set where applicable

        """
        logger.info(
            "Performing hyperparameter and model search for the best predictor models for use as benchmarkers for the "
            "regression objectives"
        )

        self.X = X
        self.Y = Y

        # Fit the ML regression model(s)
        self.objectives.fit(
            self.X,
            self.Y,
            self.variables,
            model_search_bool=True,
            test_ratio=test_ratio,
            **kwargs,
        )
        self.fitted = True

        logger.info(
            f"Calculating the approx. best result for the benchmark function. This can take several minutes to "
            f"complete"
        )

        # Calculates the best result using the benchmark model(s) with the input constraints provided (if any)
        xl = 0
        xu = 1
        pop_size = 1200
        generations = 100

        ref_dirs = get_reference_directions("energy", self.objectives.n_obj, pop_size * 0.9, seed=1)
        algorithm = UNSGA3(ref_dirs, pop_size=pop_size)
        problem = NSGABenchmarkProblemWrapper(xl, xu, self.variables, self.objectives, self.constraints)
        termination = get_termination("n_gen", generations)

        # Runs the U-NSGA-III algorithm
        res = minimize(problem, algorithm, termination, verbose=False)

        # Correct by * -1 for maximising objectives from the U-NSGA-III algorithm
        for obj_index, obj in enumerate(self.objectives.max_bool_dict):
            if self.objectives.max_bool_dict[obj]:
                if self.objectives.n_obj > 1:
                    res.F[:, obj_index] *= -1
                else:
                    res.F[obj_index] *= -1

        self.best = res.F

        logger.info(
            f"Identified the best predictor models and hyperparameters for the benchmark functions for the regression "
            f"objectives: {self.objectives.predictor_types}"
        )

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Returns the corresponding output values and standard deviations using the inputted X array and benchmark models
        and calculable objectives

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation

        Returns
        ----------
        Y: np.ndarray
            Complete Y array of output values from all calculable and regression objectives using the provided
            X array
        Y_pred_stddev: np.ndarray
            Complete array of corresponding standard deviations for the predicted Y-values of all calculable and
            regression objectives using the provided X array

        """
        Y = np.zeros((X.shape[0], self.objectives.n_obj), dtype=np.float64)
        Y_stddev = np.zeros_like(Y)
        for obj_index, obj in enumerate(self.objectives.objectives):
            if isinstance(obj, RegressionObjective):
                Y[:, obj_index], Y_stddev[:, obj_index] = obj.predict(X)
            elif isinstance(obj, CalculableObjective):
                Y[:, obj_index], Y_stddev[:, obj_index] = obj.calculate(X)

        return Y.astype("float"), Y_stddev.astype("float")

    def best_result(self) -> np.ndarray:
        """

        Returns the best result for the benchmark function

        """
        return self.best

    def save(self, dir: str) -> None:
        """

        Saves the ModelBenchmark information

        Parameters
        ----------
        dir: str
            Directory to save the pickle and Excel files of the ModelBenchmark information

        """
        if not os.path.exists(dir):
            os.makedirs(dir)

        # Temporarily stored the information in these lists
        model_info_list = []
        model_info_columns = []
        for obj in self.objectives.objectives:
            if isinstance(obj, RegressionObjective):
                if obj.obj_function.include_validation:
                    model_info_list.append(obj.obj_function.performance_metrics["Train RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["Train r2"])
                    model_info_list.append(obj.obj_function.performance_metrics["Validation RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["Validation r2"])
                    model_info_columns.append(f"{obj.name} Model Train RMSE")
                    model_info_columns.append(f"{obj.name} Model Train r2")
                    model_info_columns.append(f"{obj.name} Model Validation RMSE")
                    model_info_columns.append(f"{obj.name} Model Validation r2")
                else:
                    model_info_list.append(obj.obj_function.performance_metrics["RMSE"])
                    model_info_list.append(obj.obj_function.performance_metrics["r2"])
                    model_info_columns.append(f"{obj.name} Model Train RMSE")
                    model_info_columns.append(f"{obj.name} Model Train r2")
                model_info_list.append(obj.obj_function.name)

                if obj.obj_function.name == "gp":
                    if isinstance(obj.model_params["kernel"], AdditiveStructureKernel):
                        model_info_list.append(
                            f"AdditiveStructureKernel(ScaleKernel(MaternKernel(ard_num_dims={self.variables.n_var}, "
                            f"nu={obj.model_params['kernel'].base_kernel.base_kernel.nu}, "
                            f"lengthscale_prior=GammaPrior(3.0, 6.0)), outputscale_prior=GammaPrior(2.0, 0.15)), "
                            f"num_dims={self.variables.n_var})"
                        )
                    else:
                        model_info_list.append(
                            f"ScaleKernel(MaternKernel(nu={obj.model_params['kernel'].base_kernel.nu}, "
                            f"ard_num_dims={self.variables.n_var}, lengthscale_prior=GammaPrior(3.0, 6.0)), "
                            f"outputscale_prior=GammaPrior(2.0, 0.15))"
                        )
                else:
                    model_info_list.append(obj.model_params)

                model_info_columns.append(f"{obj.name} Model Type")
                model_info_columns.append(f"{obj.name} Model Hyperparameters")

        model_info_list.append(datetime.now().strftime("%d-%m-%Y, %H-%M-%S"))
        model_info_columns.append("Date and Time")
        # Duplicates this information for every XY row
        model_info = np.array([model_info_list] * self.X.shape[0])

        # Creates the DataFrames of the benchmark function information
        # Constructs lists of variables and objectives with their respective units where provided
        X_columns = []
        for name, units in zip(self.variables.names, self.variables.units):
            if units == "":
                X_columns.append(name)
            else:
                X_columns.append(f"{name} ({units})")

        Y_columns = []
        for name, units in zip(self.objectives.names, self.objectives.units):
            if units == "":
                Y_columns.append(name)
            else:
                Y_columns.append(f"{name} ({units})")

        benchmark_df = pd.DataFrame(
            np.hstack((self.X, self.Y, model_info)),
            columns=X_columns + Y_columns + model_info_columns,
        )

        # Writes the optimisation information DataFrames into Excel files
        with pd.ExcelWriter(
            os.path.join(
                dir,
                f'Model Benchmark {datetime.now().strftime("%d-%m-%Y, %H-%M-%S")}.xlsx',
            )
        ) as writer:
            benchmark_df.to_excel(writer, sheet_name="Model Benchmark")
            for obj in self.objectives.objectives:
                if isinstance(obj, RegressionObjective):
                    if obj.obj_function.include_validation:
                        obj.obj_function.performance_metrics["Train Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Train Parity Data"
                        )
                        obj.obj_function.performance_metrics["Val Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Val Parity Data"
                        )
                    else:
                        obj.obj_function.performance_metrics["Parity Data"].to_excel(
                            writer, sheet_name=f"{obj.name} Train Parity Data"
                        )

        # Writes a pickle file containing the model benchmark information
        benchmark_dict = {
            "Date and Time": datetime.now().strftime("%d-%m-%Y, %H-%M-%S"),
            "Benchmark Table": benchmark_df,
            "X": self.X,
            "Y": self.Y,
        }
        for obj in self.objectives.objectives:
            if isinstance(obj, RegressionObjective):
                benchmark_dict[f"{obj.name} Model Type"] = obj.obj_function.name
                if obj.obj_function.name == "gp":
                    if isinstance(obj.model_params["kernel"], AdditiveStructureKernel):
                        benchmark_dict[f"{obj.name} Model Hyperparameters"] = (
                            f"AdditiveStructureKernel(ScaleKernel(MaternKernel(ard_num_dims={self.variables.n_var}, "
                            f"nu={obj.model_params['kernel'].base_kernel.base_kernel.nu}, "
                            f"lengthscale_prior=GammaPrior(3.0, 6.0)), outputscale_prior=GammaPrior(2.0, 0.15)), "
                            f"num_dims={self.variables.n_var})"
                        )
                    else:
                        benchmark_dict[f"{obj.name} Model Hyperparameters"] = (
                            f"ScaleKernel(MaternKernel(nu={obj.model_params['kernel'].base_kernel.nu}, "
                            f"ard_num_dims={self.variables.n_var}, lengthscale_prior=GammaPrior(3.0, 6.0)), "
                            f"outputscale_prior=GammaPrior(2.0, 0.15))"
                        )
                else:
                    benchmark_dict[f"{obj.name} Model Hyperparameters"] = obj.model_params
                benchmark_dict[f"{obj.name} Model Performance Metrics"] = obj.obj_function.performance_metrics

        with open(
            os.path.join(
                dir,
                f"Model Benchmark {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.pickle",
            ),
            "wb",
        ) as file:
            pickle.dump(benchmark_dict, file, protocol=pickle.HIGHEST_PROTOCOL)


class NSGABenchmarkProblemWrapper(Problem):
    """

    Class for defining a problem in the ModelBenchmark class

    Parameters
    ----------
    xl: int | float | List[int | float]
        Lower bounds for the variables for the NSGA algorithm. If integer, all lower bounds are equal to the integer
        value
    xu: int | float | List[int | float]
        Upper bounds for the variables for the NSGA algorithm. If integer, all upper bounds are equal to the integer
        value
    variables: VariablesList
        VariablesList object that contains information about all variables
    objectives: ObjectivesList
        ObjectivesList object that contains information about all objectives
    constraints: ConstraintsList
        ConstraintsList object that contains information about all constraints

    """

    def __init__(
        self,
        xl: Union[int, float, List[Union[int, float]]],
        xu: Union[int, float, List[Union[int, float]]],
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: ConstraintsList,
    ):
        self.variables = variables
        self.objectives = objectives
        self.constraints = constraints

        n_constr = self.constraints.n_constr if constraints is not None else 0

        super().__init__(
            n_var=self.variables.n_var,
            n_obj=self.objectives.n_obj,
            n_constr=n_constr,
            xl=xl,
            xu=xu,
        )

    def _evaluate(self, X: np.ndarray, out: Dict[str, np.ndarray], **kwargs) -> None:
        """

        Function to calculate the objective(s) and constraint(s) (if any) values

        Parameters
        ----------
        X: np.ndarray
            NSGA generated candidates' X-values, which are min-max normalised
        out: Dict[str, np.ndarray]
            Contains the objective(s) and constraint(s) (if any) values for the generated candidates

        """
        # Un-transform the min-max normalised X array
        X = (X * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Infer the objective values by drawing samples from the distribution at the X-values
        if self.variables.num_cat_descriptor_var > 0:
            F, _ = self.objectives.evaluate(self.variables.descriptor_to_name(X))
        else:
            F, _ = self.objectives.evaluate(X)

        # * -1 for objectives that are maximising
        for obj_index, obj in enumerate(self.objectives.max_bool_dict):
            if self.objectives.max_bool_dict[obj]:
                F[:, obj_index] *= -1
        out["F"] = F

        # Evaluate the pymoo constraints
        if self.n_constr > 0:
            # Min-max normalises the X array
            X = (X - np.array(self.variables.lower_bounds)) / (
                np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds)
            )
            G = np.zeros((X.shape[0], self.n_constr))
            for index, constr in enumerate(self.constraints.constraints):
                G[:, index] = constr.evaluate_pymoo_constraint(X)
            out["G"] = G



# Based on forest-confidence-interval: Confidence intervals for Forest algorithms, Kivan Polimis, Ariel Rokem, Bryna
# Hazelton, The University of Washington, https://github.com/scikit-learn-contrib/forest-confidence-interval

from __future__ import annotations

import os
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple

import forestci as fci
import numpy as np
from hyperopt import STATUS_OK, Trials, fmin, hp, tpe
from hyperopt.pyll.base import scope
from sklearn.ensemble import RandomForestRegressor

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.perf_metrics as pm
from nemo_bo.models.base.base_model import Base_Model

if TYPE_CHECKING:
    from nemo_bo.opt.objectives import RegressionObjective
    from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


class RFModel(Base_Model):
    """

    Used to create a Random Forest model for the RegressionObjective

    See docstring for Base_Model __init__ function for information

    """

    def __init__(
        self, variables: VariablesList, objective: RegressionObjective, always_hyperparam_opt: bool = True, **kwargs
    ):
        super().__init__(variables, objective, always_hyperparam_opt)
        self.default_X_transform_type = "none"
        self.default_Y_transform_type = "none"
        self.include_validation = False
        self.name = "rf"

    def fit(self, X: np.ndarray, Y: np.ndarray, **params) -> None:
        """

        Function that is called to start the Random Forest model fitting procedure

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        params: Dict[str, Any]
            Dictionary that contains the hyperparameters for the model

        """
        if self.variables.num_cat_var > 0:
            self.X_train = self.variables.categorical_transform(X).astype("float")
        else:
            self.X_train = X

        self.X_train, self.Y_train = self.transform_by_predictor_type(self.X_train, Y=Y)

        self.fit_model(params)

        self.Y_pred, self.Y_pred_stddev = self.predict(X)
        self.performance_metrics = pm.all_performance_metrics(Y, self.Y_pred)

        self.Y_pred_error = 1.96 * self.Y_pred_stddev

    def fit_model(self, params: Dict[str, Any]) -> None:
        """

        Called by the fit and cv-related functions for fitting the Random Forest model using pre-processed X and Y
        training data in the class instance

        """
        for key in params:
            if key == "max_depth":
                params["max_depth"] = int(params["max_depth"])
            if key == "n_estimators":
                params["n_estimators"] = int(params["n_estimators"])

        # Create RandomForestRegressor
        self.model = RandomForestRegressor(random_state=1, **params)
        self.model.fit(self.X_train, self.Y_train.flatten())

    def predict(self, X: np.ndarray, X_transform: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """

        Determines the model output and its standard deviation of the X array passed into the function

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation
        X_transform: bool, Default = True
            For the default models provided natively in NEMO, the initial transformation of categorical variables and
            transformations related to standardisation/normalisation can be switched off by passing X_transform = False

        Returns
        ----------
        Y_pred: np.ndarray
            The predictive mean of the objective calculated using the model and the inputted X array. The outputted
            values have been un-transformed
        Y_pred_stddev: np.ndarray
            The predictive standard deviation of the objective calculated using the model and the inputted X array. The
            outputted values have been un-transformed

        """
        # X needs to be a 2D array
        if X.ndim == 1:
            X = X.reshape(1, -1)

        if X_transform:
            if self.variables.num_cat_var > 0:
                X = self.variables.categorical_transform(X).astype("float")
            X = self.transform_only_by_predictor_type(X)

        Y_mean = self.model.predict(X.astype(float))

        Y_var = np.absolute(
            fci.random_forest_error(self.model, self.X_train, X, calibrate=False if X.shape[0] <= 20 else True)
        )
        Y_pred_upper = Y_mean + (1.96 * np.sqrt(Y_var))

        Y_pred = self.objective.inverse_transform(Y_mean)
        Y_pred_upper = self.objective.inverse_transform(Y_pred_upper)
        Y_pred_stddev = (np.subtract(Y_pred_upper, Y_pred)) / 1.96

        return Y_pred.flatten(), Y_pred_stddev.flatten()

    def draw_samples(self, X: np.ndarray, X_transform: bool = True) -> np.ndarray:
        """

        Generates samples from a normal distribution using the X array

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation
        X_transform: bool, Default = True
            For the default models provided natively in NEMO, the initial transformation of categorical variables and
            transformations related to standardisation/normalisation can be switched off by passing X_transform = False

        Returns
        ----------
        np.ndarray
            Drawn samples of the objective using the model and the inputted X array. The outputted values have been
            un-transformed

        """
        Y_pred, Y_pred_stddev = self.predict(X, X_transform=X_transform)

        return np.random.default_rng().normal(Y_pred, Y_pred_stddev)

    @staticmethod
    def default_params():
        """

        Function that returns the default hyperparameters to use for the hyperparameter optimisation of Random Forest
        models using the hyperopt package. Returns a dictionary that is used for the space argument in the hyperopt.fmin
        function

        """
        return {
            "max_depth": scope.int(hp.quniform("max_depth", 5, 100, 1)),
            "min_samples_leaf": hp.uniform("min_samples_leaf", 0.01, 0.5),
            "min_samples_split": hp.uniform("min_samples_split", 0.01, 0.75),
            "n_estimators": scope.int(hp.quniform("n_estimators", 40, 2000, 20)),
        }

    def hyperparam_opt(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float,
        predictor_params_dict: Optional[Dict[str, Dict[str, Callable]]] = None,
        **kwargs,
    ) -> Tuple[RFModel, Dict[str, Any]]:
        """

        Function to be called for performing hyperparameter optimisation for Random Forest models using the hyperopt
        package. Returns the model fitted with the best hyperparameters and a dictionary containing the best
        hyperparameters

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float
            The proportion of inputted X and Y arrays to be split for the validation set where applicable
        predictor_params_dict: Dict[str, Dict[str, Callable]], Default = None
            Dictionary with a key 'rf' and a corresponding value that is a dictionary that is used for the space
            argument in the hyperopt.fmin function

        Returns
        -------
        model: RFModel
            The model fitted with the best hyperparameters found using the hyperopt package

        model_params: Dict[str, Any]
            Dictionary containing the best hyperparameters found using the hyperopt package

        """
        if predictor_params_dict is None:
            predictor_params_dict = self.default_params()
        else:
            predictor_params_dict = predictor_params_dict.get(self.name, self.default_params())

        # if self.objective.hyperopt_evals is None:
        #     hyperopt_evals = 1
        # else:
        #     hyperopt_evals = self.objective.hyperopt_evals.get(self.name, 1)
        if self.objective.hyperopt_evals is None:
            hyperopt_evals = 30
        else:
            hyperopt_evals = self.objective.hyperopt_evals.get(f"{self.name}", 30)

        def func(X: np.ndarray, Y: np.ndarray, model_params: Dict[str, Any]) -> Dict[str, Any]:
            cv_results = self.cv(X, Y, model_params, test_ratio=test_ratio, **kwargs)

            return {
                "loss": cv_results["Mean Test RMSE"],
                "status": STATUS_OK,
            }

        trials = Trials()
        model_params = fmin(
            fn=partial(func, X, Y),
            space=predictor_params_dict,
            algo=tpe.suggest,
            max_evals=hyperopt_evals,
            trials=trials,
        )

        model = self.new_instance(self.__class__, **kwargs)
        model.fit(X, Y, test_ratio=test_ratio, **model_params, **kwargs)

        logger.info(f"Completed hyperparameter opt")
        return model, model_params



from __future__ import annotations

import math
import os
from datetime import datetime
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from hyperopt import STATUS_OK, Trials, fmin, hp, tpe
from hyperopt.pyll.base import scope
from keras import backend as K
from keras import callbacks
from keras.layers import Dense, Input, InputSpec, Wrapper, concatenate
from keras.models import Model
from keras.optimizers import Optimizer
from sklearn.model_selection import train_test_split
from tensorflow import keras

tf.keras.backend.set_floatx("float64")

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.perf_metrics as pm
from nemo_bo.models.base.base_model import Base_Model
from nemo_bo.models.base.nn_save_checkpoint import SaveModelCheckPoint
from nemo_bo.utils.data_proc import sort_train_test_split_shuffle

if TYPE_CHECKING:
    from nemo_bo.opt.objectives import RegressionObjective
    from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


tfpl = tfp.layers
tfd = tfp.distributions

gpus = tf.config.list_physical_devices("GPU")
if gpus:
    try:
        # Memory growth needs to be the same across GPUs
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)
        logical_gpus = tf.config.list_logical_devices("GPU")
        print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
    except RuntimeError as e:
        print(e)


class ConcreteDropout(Wrapper):
    """

    Code copied from:
    João Caldeira, Brian Nord, Deeply Uncertain: Comparing Methods of Uncertainty Quantification in Deep Learning
    Algorithms, arXiv:2004.10710, https://github.com/deepskies/DeeplyUncertain-Public

    """

    def __init__(
        self,
        layer,
        weight_regularizer=0,
        dropout_regularizer=1e-5,
        init_min=0.1,
        init_max=0.1,
        is_mc_dropout=True,
        **kwargs,
    ):
        assert "kernel_regularizer" not in kwargs
        super(ConcreteDropout, self).__init__(layer, **kwargs)
        self.weight_regularizer = weight_regularizer
        self.dropout_regularizer = dropout_regularizer
        self.is_mc_dropout = is_mc_dropout
        self.supports_masking = True
        self.p_logit = None
        self.init_min = np.log(init_min) - np.log(1.0 - init_min)
        self.init_max = np.log(init_max) - np.log(1.0 - init_max)

    def build(self, input_shape=None):
        self.input_spec = InputSpec(shape=input_shape, dtype=tf.float64)
        if not self.layer.built:
            self.layer.build(input_shape)
            self.layer.built = True
        super(ConcreteDropout, self).build()

        # initialise p
        self.p_logit = self.add_weight(
            name="p_logit",
            shape=(1,),
            initializer=tf.random_uniform_initializer(self.init_min, self.init_max),
            dtype=tf.dtypes.float64,
            trainable=True,
        )

    def compute_output_shape(self, input_shape):
        return self.layer.compute_output_shape(input_shape)

    def concrete_dropout(self, x, p):
        eps = 1e-07
        temp = 0.1

        unif_noise = tf.random.uniform(shape=tf.shape(x), dtype=tf.float64)
        drop_prob = (
            tf.math.log(p + eps)
            - tf.math.log(1.0 - p + eps)
            + tf.math.log(unif_noise + eps)
            - tf.math.log(1.0 - unif_noise + eps)
        )
        drop_prob = tf.math.sigmoid(drop_prob / temp)
        random_tensor = 1.0 - drop_prob

        retain_prob = 1.0 - p
        x *= random_tensor
        x /= retain_prob
        return x

    def call(self, inputs, training=None):
        p = tf.math.sigmoid(self.p_logit)

        # initialise regulariser / prior KL term
        input_dim = inputs.shape[-1]  # last dim
        weight = self.layer.kernel
        kernel_regularizer = self.weight_regularizer * tf.reduce_sum(tf.square(weight)) / (1.0 - p)
        dropout_regularizer = p * tf.math.log(p) + (1.0 - p) * tf.math.log(1.0 - p)
        dropout_regularizer *= self.dropout_regularizer * input_dim
        regularizer = tf.reduce_sum(kernel_regularizer + dropout_regularizer)
        if self.is_mc_dropout:
            return self.layer.call(self.concrete_dropout(inputs, p)), regularizer
        else:

            def relaxed_dropped_inputs():
                return self.layer.call(self.concrete_dropout(inputs, p)), regularizer

            return (
                tf.keras.backend.in_train_phase(relaxed_dropped_inputs, self.layer.call(inputs), training=training),
                regularizer,
            )


class NNConcreteDropoutModel(Base_Model):
    """

    Used to create a neural network model with self-learned dropout probability for the RegressionObjective

    See docstring for Base_Model __init__ function for information

    """

    def __init__(
        self,
        variables: VariablesList,
        objective: RegressionObjective,
        always_hyperparam_opt: bool = False,
        es_patience: int = 300,  # normally 300
    ):
        super().__init__(variables, objective, always_hyperparam_opt)
        self.default_X_transform_type = "standardisation"
        self.default_Y_transform_type = "standardisation"
        self.include_validation = True
        self.name = "nn_concrete"

        # Creates the callback method to stop fitting before the all epochs are completed based on the prediction
        # accuracy of the validation set
        self.earlystopping = callbacks.EarlyStopping(
            monitor="val_loss",
            min_delta=0.001,
            patience=es_patience,
            verbose=0,
            mode="min",
        )

        # The directory to save the model parameters when it detects a model with better prediction accuracy of the
        # validation set
        self.dirname_checkpoint = os.path.join(os.getcwd(), "ML Models", f"{self.name}", "Checkpoints")
        if not os.path.exists(self.dirname_checkpoint):
            os.makedirs(self.dirname_checkpoint)

    @staticmethod
    def mse_loss(true, pred):
        """

        Function to calculate the mean squared error (MSE) during training as a monitoring metric

        """
        n_outputs = pred.shape[1] // 2
        mean = pred[:, :n_outputs]
        return tf.reduce_mean((true - mean) ** 2, -1)

    @staticmethod
    def heteroscedastic_loss(true, pred):
        """

        Function to calculate the heteroscedastic uncertainty during training as the loss function

        """
        n_outputs = pred.shape[1] // 2
        mean = pred[:, :n_outputs]
        log_var = pred[:, n_outputs:]
        precision = tf.math.exp(-log_var)
        return tf.reduce_sum(precision * (true - mean) ** 2.0 + log_var, -1)

    def structure(
        self,
        learning_rate: float,
        hidden_units: int,
        hidden_layers: int,
        act_func: str,
        optimizer: Optimizer,
        wd: float = 0.0,
    ) -> Model:
        """

        Function that defines the structure of the input, hidden layers, and output for a neural network model with
        self-learned dropout probability

        Parameters
        ----------
        learning_rate: float
            Assigns the learning rate for the optimiser
        hidden_units: int
            The number of hidden units within every hidden layer. Currently, all hidden layers will have this number
            of hidden units
        hidden_layers: int
            The number of hidden layers in the neural network
        dropout_proba: float
            The dropout probability to be applied to the hidden layers
        act_func: str
            The activation function to apply on every hidden layer
        optimizer: Optimizer
            The type of optimiser to use to fit the model
        wd: float
            The weight regulariser to use in the model

        Returns
        -------
        model:
            The neural network model that can learn the dropout probability

        """
        K.clear_session()
        tf.random.set_seed(1)
        tf.random.uniform([1], seed=1)
        dropout_reg = 2 / self.X_train.shape[0]
        losses = []
        inputs = Input(shape=(self.X_train.shape[1],))
        x = inputs
        for i in range(hidden_layers):
            x, loss = ConcreteDropout(
                Dense(hidden_units, activation=act_func),
                weight_regularizer=wd,
                dropout_regularizer=dropout_reg,
            )(x)
            losses.append(loss)
        mean, loss = ConcreteDropout(
            Dense(self.Y_train.shape[1]),
            weight_regularizer=wd,
            dropout_regularizer=dropout_reg,
        )(x)
        losses.append(loss)
        log_var, loss = ConcreteDropout(
            Dense(self.Y_train.shape[1]),
            weight_regularizer=wd,
            dropout_regularizer=dropout_reg,
        )(x)
        losses.append(loss)
        out = concatenate([mean, log_var])
        model = Model(inputs, out)

        for loss in losses:
            model.add_loss(loss)

        # model.summary()

        model.compile(
            optimizer=optimizer(learning_rate=learning_rate),
            loss=self.heteroscedastic_loss,
            metrics=[self.mse_loss],
        )
        assert len(model.layers[1].trainable_weights) == 3

        return model

    def fit(
        self, X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, sort_before_split: bool = None, **params
    ) -> None:
        """

        Function that is called to start the concrete dropout neural network model fitting procedure

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation set where applicable
        sort_before_split: bool, Default = True,
            Whether the inputted X and Y arrays should be sorted such that the Y-values are in ascending order before
            splitting into the training and test sets (and validation sets if applicable)
        params: Dict[str, Any]
            Dictionary that contains the hyperparameters for the model

        """
        if sort_before_split:
            (
                X_train,
                X_val,
                Y_train,
                Y_val,
            ) = sort_train_test_split_shuffle(X, Y, test_ratio=test_ratio, seed=1)
        else:
            X_train, X_val, Y_train, Y_val = train_test_split(X, Y, test_size=test_ratio, random_state=1)

        Y_train = Y_train.astype("float")
        Y_val = Y_val.astype("float")

        self.X_train, self.X_val, self.Y_train, self.Y_val = (
            X_train,
            X_val,
            Y_train,
            Y_val,
        )
        if self.variables.num_cat_var > 0:
            self.X_train = self.variables.categorical_transform(self.X_train).astype("float")
            self.X_val = self.variables.categorical_transform(self.X_val).astype("float")

        self.X_train, self.Y_train = self.transform_by_predictor_type(self.X_train, Y=self.Y_train)
        self.X_val, self.Y_val = self.transform_only_by_predictor_type(self.X_val, self.Y_val)

        self.fit_model(params)

        self.Y_train_pred, self.Y_train_pred_stddev = self.predict(X_train)
        self.Y_val_pred, self.Y_val_pred_stddev = self.predict(X_val)
        self.performance_metrics = pm.all_performance_metrics_train_val(
            Y_train, self.Y_train_pred, Y_val, self.Y_val_pred
        )

        self.Y_train_error = 1.96 * self.Y_train_pred_stddev
        self.Y_val_error = 1.96 * self.Y_val_pred_stddev

    def fit_model(self, params: Dict[str, Any]) -> None:
        """

        Called by the fit and cv-related functions for fitting the neural network model with self-learned dropout
        probability using pre-processed X and Y training data in the class instance

        """
        hidden_units = int(params.get("hidden_units", self.X_train.shape[1] + 1))
        hidden_layers = int(params.get("hidden_layers", 2))
        batch_size = int(params.get("batch_size", 32))
        learning_rate = params.get("learning_rate", 0.001)
        max_epochs = int(params.get("max_epochs", 4000))
        act_func = params.get("act_func", "relu")
        optimizer = params.get("optimizer", keras.optimizers.Adam)

        self.filepath = os.path.join(
            self.dirname_checkpoint,
            f"Checkpoint, Concrete Dropout, {datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}.h5",
        )

        self.model = self.structure(
            learning_rate,
            hidden_units,
            hidden_layers,
            act_func,
            optimizer,
        )
        self.savemodelcheckpoint = SaveModelCheckPoint(self.model, self.filepath)
        self.history = self.model.fit(
            self.X_train,
            self.Y_train,
            validation_data=(self.X_val, self.Y_val),
            epochs=max_epochs,
            verbose=0,
            batch_size=batch_size,
            callbacks=[
                callbacks.LambdaCallback(on_epoch_end=self.savemodelcheckpoint.save_weights),
                self.earlystopping,
            ],
        )

        self.model = self.structure(
            learning_rate,
            hidden_units,
            hidden_layers,
            act_func,
            optimizer,
        )
        self.model.load_weights(self.filepath)

    def predict(self, X: np.ndarray, X_transform: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """

        Determines the model output and its standard deviation of the X array passed into the function

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation
        X_transform: bool, Default = True
            For the default models provided natively in NEMO, the initial transformation of categorical variables and
            transformations related to standardisation/normalisation can be switched off by passing X_transform = False

        Returns
        ----------
        Y_pred: np.ndarray
            The predictive mean of the objective calculated using the model and the inputted X array. The outputted
            values have been un-transformed
        Y_pred_stddev: np.ndarray
            The predictive standard deviation of the objective calculated using the model and the inputted X array. The
            outputted values have been un-transformed

        """
        # X needs to be a 2D array
        if X.ndim == 1:
            X = X.reshape(1, -1)

        if X_transform:
            if self.variables.num_cat_var > 0:
                X = self.variables.categorical_transform(X).astype("float")
            X = self.transform_only_by_predictor_type(X)

        observed_Y_pred = np.array([self.model.predict(X) for _ in range(10)])
        Y_pred = observed_Y_pred[:, :, :1]
        Y_pred_stddev = np.sqrt(np.exp(observed_Y_pred[:, :, 1:]))

        # Code adapted from:
        #     1. João Caldeira, Brian Nord, Deeply Uncertain: Comparing Methods of Uncertainty Quantification in Deep
        #     Learning Algorithms, arXiv:2004.10710, https://github.com/deepskies/DeeplyUncertain-Public,
        #     2. Yarin Gal, Jiri Hron, Alex Kendall, Concrete Dropout, arXiv:1705.07832,
        #     https://github.com/yaringal/ConcreteDropout
        Y_val_epistemic_unc = np.std(np.array(Y_pred), axis=0)
        Y_val_aleatoric_unc = np.sqrt(np.mean(np.array(Y_pred_stddev) * np.array(Y_pred_stddev), axis=0))
        Y_val_total_unc = np.sqrt(Y_val_aleatoric_unc**2 + Y_val_epistemic_unc**2)

        Y_pred = np.mean(Y_pred, axis=0)
        Y_pred_upper = Y_pred + (1.96 * Y_val_total_unc)

        Y_pred = self.objective.inverse_transform(Y_pred)
        Y_pred_upper = self.objective.inverse_transform(Y_pred_upper)
        Y_pred_stddev = (np.subtract(Y_pred_upper, Y_pred)) / 1.96

        return Y_pred.flatten(), Y_pred_stddev.flatten()

    def draw_samples(self, X: np.ndarray, X_transform: bool = True) -> np.ndarray:
        """

        Generates samples from a normal distribution using the X array

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation
        X_transform: bool, Default = True
            For the default models provided natively in NEMO, the initial transformation of categorical variables and
            transformations related to standardisation/normalisation can be switched off by passing X_transform = False

        Returns
        ----------
        np.ndarray
            Drawn samples of the objective using the model and the inputted X array. The outputted values have been
            un-transformed

        """
        Y_pred, Y_pred_stddev = self.predict(X, X_transform=X_transform)

        return np.random.default_rng().normal(Y_pred, Y_pred_stddev)

    @staticmethod
    def default_params(X: np.ndarray):
        """

        Function that returns the default hyperparameters to use for the hyperparameter optimisation of a neural
        network model that can self-learn dropout probability using the hyperopt package. Returns a dictionary that is
        used for the space argument in the hyperopt.fmin function

        """
        return {
            "learning_rate": hp.uniform("learning_rate", 0.001, 0.01),
            "hidden_units": scope.int(hp.quniform("hidden_units", math.ceil(X.shape[1] / 2), (X.shape[1] + 1), 1)),
            "hidden_layers": scope.int(hp.quniform("hidden_layers", 1, 2, 1)),
            "batch_size": scope.int(hp.quniform("batch_size", math.ceil(X.shape[0] / 10), X.shape[0], 1)),
        }

    def hyperparam_opt(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float,
        sort_before_split: bool = True,
        predictor_params_dict: Optional[Dict[str, Dict[str, Callable]]] = None,
        **kwargs,
    ) -> Tuple[NNConcreteDropoutModel, Dict[str, Any]]:
        """

        Function to be called for performing hyperparameter optimisation for neural network models with self-learned
        dropout probability using the hyperopt package. Returns the model fitted with the best hyperparameters and a
        dictionary containing the best hyperparameters

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float
            The proportion of inputted X and Y arrays to be split for the validation set where applicable
        sort_before_split: bool, Default = True,
            Whether the inputted X and Y arrays should be sorted such that the Y-values are in ascending order before
            splitting into the training and test sets (and validation sets if applicable)
        predictor_params_dict: Dict[str, Dict[str, Callable]], Default = None
            Dictionary with a key 'nn_concrete' and a corresponding value that is a dictionary that is used for the
            space argument in the hyperopt.fmin function

        Returns
        -------
        model: NNBayesianModel
            The model fitted with the best hyperparameters found using the hyperopt package

        model_params: Dict[str, Any]
            Dictionary containing the best hyperparameters found using the hyperopt package

        """
        if predictor_params_dict is None:
            predictor_params_dict = self.default_params(X)
        else:
            predictor_params_dict = predictor_params_dict.get(self.name, self.default_params(X))

        # if self.objective.hyperopt_evals is None:
        #     hyperopt_evals = 1
        # else:
        #     hyperopt_evals = self.objective.hyperopt_evals.get(self.name, 1)
        if self.objective.hyperopt_evals is None:
            hyperopt_evals = 20
        else:
            hyperopt_evals = self.objective.hyperopt_evals.get(f"{self.name}", 20)

        def func(X: np.ndarray, Y: np.ndarray, model_params: Dict[str, Any]) -> Dict[str, Any]:
            loss = self.non_cv_train_val_test(
                X, Y, model_params, test_ratio=test_ratio, sort_before_split=sort_before_split, **kwargs
            )

            return {
                "loss": loss,
                "status": STATUS_OK,
            }

        trials = Trials()
        model_params = fmin(
            fn=partial(func, X, Y),
            space=predictor_params_dict,
            algo=tpe.suggest,
            max_evals=hyperopt_evals,
            trials=trials,
        )

        model = self.new_instance(self.__class__, **kwargs)
        model.fit(X, Y, test_ratio=test_ratio, **model_params, **kwargs)

        logger.info(f"Completed hyperparameter opt")
        return model, model_params



import copy
import os

from attrs import define, field
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.perf_metrics as pm
import nemo_bo.utils.plotter as plotter
from nemo_bo.models.base.available_models import create_predictor_list
from nemo_bo.models.base.base_model import Base_Model
from nemo_bo.models.nn_bayesian import NNBayesianModel
from nemo_bo.models.nn_concrete import NNConcreteDropoutModel
from nemo_bo.opt.variables import VariablesList
from nemo_bo.utils.data_proc import sort_train_test_split_shuffle, remove_nan

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define(kw_only=True)
class Objective:
    """

    Base class for an objective

    Parameters
    ----------
    name: str
        Name of the objective
    obj_max_bool: bool
        Whether the objective is to be maximised (True) or minimised (False)
    lower_bound: int | float
        The lower bound of the objective
    upper_bound: int | float
        The upper bound of the objective
    transformer: Optional. Default = None
        The transformation that will be applied to the objective for modelling. When left as the default
        None, the modelling uses the default transformations specified by the model type. Can be specified as the
        strings "normalisation", "standardisation", or "none" for min-max normalisation, standardisation to a mean of 0
        and a standard deviation of 1, or no transformation respectively. Alternatively, a class instance with
        fit_transform, transform, and inverse_transform functions, such as a scikit-learn transformer can be provided
        directly too.
    units: str, Default = ""
        The units for the objective

    """

    name: str
    obj_max_bool: bool
    lower_bound: Union[int, float]
    upper_bound: Union[int, float]
    transformer: Optional = None
    units: str = ""

    def transform(self, Y: np.ndarray) -> np.ndarray:
        """

        Fits a transformation scalar to the inputted Y array when self.transformer is "normalisation",
        "standardisation" or a transformer object. The Y array is subsequently transformed according to the
        fitted scalar. When self.transformer = "none", no transformation occurs

        Parameters
        ----------
        Y: np.ndarray
            Y array containing untransformed values for one objective

        """
        if self.transformer is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, 2. Set transformer = 'none', 'normalisation', or "
                "'standarisation' when creating the variable instance, or 3. Pass a transformer such as a scikit-learn "
                "transformer when creating the variable instance"
            )

        if Y.ndim == 1:
            Y = Y.reshape(-1, 1)

        if self.transformer == "none":
            return Y
        elif self.transformer == "normalisation":
            self.transformer = MinMaxScaler()
        elif self.transformer == "standardisation":
            self.transformer = StandardScaler()

        return self.transformer.fit_transform(Y)

    def transform_only(self, Y: np.ndarray) -> np.ndarray:
        """

        The Y array is transformed according to the fit scalar in the self.transformer object. When
        self.transformer = "none", no transformation occurs

        Parameters
        ----------
        Y: np.ndarray
            Y array containing untransformed values for one objective

        """

        if Y.ndim == 1:
            Y = Y.reshape(-1, 1)

        if self.transformer is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, 2. Set transformer = 'none', 'normalisation', or "
                "'standarisation' when creating the variable instance, or 3. Pass a transformer such as a scikit-learn "
                "transformer when creating the variable instance"
            )

        if self.transformer == "none":
            return Y
        else:
            return self.transformer.transform(Y)

    def inverse_transform(self, Y_transform: np.ndarray) -> np.ndarray:
        """

        The Y_transform array undergoes an inverse transform according to the fit scalar in the self.transformer
        object. When self.transformer = "none", no transformation occurs

        Parameters
        ----------
        Y_transform: np.ndarray
            Y array containing transformed values for one objective

        """

        if Y_transform.ndim == 1:
            Y_transform = Y_transform.reshape(-1, 1)

        if self.transformer == "none":
            return Y_transform
        else:
            return self.transformer.inverse_transform(Y_transform)


@define(kw_only=True)
class RegressionObjective(Objective):
    """

    Class for an objective to be modelled using a regression machine learning model

    Parameters
    ----------
    name: str
        Name of the objective
    obj_max_bool: bool
        Whether the objective is to be maximised (True) or minimised (False)
    lower_bound: int | float
        The lower bound of the objective
    upper_bound: int | float
        The upper bound of the objective
    transformer: Optional. Default = None
        The transformation that should be applied to the objective for modelling. When left as the default
        None, the modelling uses the default transformations specified by the model type. Can be specified as the
        strings "normalisation", "standardisation", or "none" for min-max normalisation, standardisation to a mean of 0
        and a standard deviation of 1, or no transformation respectively. Alternatively, a class instance with
        fit_transform, transform, and inverse_transform functions, such as a scikit-learn transformer can be provided
        directly too.
    units: str, Default = ""
        The units for the objective
    predictor_type: str | Base_Model | List[str | Base_Model], Default = None
        Specifies the type of predictor/model to use for the objective. If unspecified, it will use all of the default
        predictors/models built-in NEMO
    predictor_params_dict: Dict[str, Dict[str, Callable]], Default = None
        Specifies the predictor/model hyperparameters to optimise during model fitting using the HyperOpt package.
        Dictionary with keys corresponding to the model names, and corresponding values that are dictionaries that
        specify the hyperameters to optimise in the HyperOpt format. Any predictor/model types that are unspecified in
        this dictionary will use the default hyperparameters defined by each the model
    gp_kernel_choices: list, Default = None
        Specifies the Gaussian Process (GP) kernels to use. Will only use this argument if GP's are used for model
        fitting. If unspecified, will use the default kernels defined for GP models in NEMO (can find these in
        nemo_bo.models.gp)
    model_params: Dict[str, Any], Default = None
        Specifies the model hyperparameters to use for model fitting
    hyperopt_evals: Dict[str, int], Default = None
        Overrides the default number of HyperOpt iterations for each specified model's hyperparameter optimisation.
        The dictionary keys correspond to the model name, and the values correspond to the number of iterations to use.
        Any predictor/model types that are unspecified in this dictionary will use the default number of iterations
        defined by each the model

    """

    predictor_type: Optional[Union[str, Base_Model, List[Union[str, Base_Model]]]] = None
    predictor_params_dict: Optional[Dict[str, Dict[str, Callable]]] = None
    gp_kernel_choices: Optional[list] = None
    model_params: Optional[Dict[str, Any]] = None
    hyperopt_evals: Optional[Dict[str, int]] = None
    input_predictor_type: str = field(init=False)
    obj_function: Base_Model = field(init=False)
    permutation_feature_importance: pd.Series = field(init=False)

    def __attrs_post_init__(self):
        self.input_predictor_type = self.predictor_type = create_predictor_list(self.predictor_type)
        self.obj_function = None
        self.permutation_feature_importance = None

    def fit_regressor(
        self, X: np.ndarray, Y: np.ndarray, variables: VariablesList, test_ratio: Optional[float] = 0.2, **kwargs
    ) -> None:
        """

        Fits regression objectives to a specific predictor/model type (self.predictor_type) and defined model
        hyperparameters (self.model_params)

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        variables: VariablesList
            VariablesList object that contains information about all variables
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation and test sets where applicable

        """
        X, Y = remove_nan(X, Y)

        self.obj_function = self.predictor_type(variables, self)
        self.obj_function.fit(X, Y, test_ratio=test_ratio, **self.model_params, **kwargs)

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Returns the untransformed predicted mean and standard deviation using the fitted machine learning models at
        the X values

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        Returns
        -------
        Y_pred: np.ndarray
            The untransformed predicted mean values using the fitted machine learning models at the X values
        Y_pred_stddev: np.ndarray
            The untransformed predicted standard deviations using the fitted machine learning models at the X values

        """
        Y_pred, Y_pred_stddev = self.obj_function.predict(X)

        return Y_pred.astype("float"), Y_pred_stddev.astype("float")

    def draw_samples(self, X: np.ndarray) -> np.ndarray:
        """

        Generates samples from a normal distribution using the X array

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        Returns
        -------
        np.ndarray
            Drawn samples using the model and the inputted X array. The outputted values have been un-transformed

        """

        return self.obj_function.draw_samples(X).astype("float")

    def model_and_hyperparameter_opt(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        model_search_bool: bool,
        test_ratio: Optional[float] = None,
        test_data: Optional[List[np.ndarray]] = None,
        **kwargs,
    ) -> None:
        """

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        variables: VariablesList
            VariablesList object that contains information about all variables
        model_search_bool: bool
            Whether automated model and hyperparameter optimisation is to be performed during fitting the regression
            models
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation and test sets where applicable
        test_data: List[np.ndarray], Default = None,
            Test data to use for the initial model and hyperparameter optimisation, instead of creating it
            automatically from the provided X and Y arrays. The test X array is at index 0 and the test Y array is at
            index 1

        """
        X, Y = remove_nan(X, Y)

        if model_search_bool:
            test_loss_list = []
            model_list = []
            model_params_list = []

            if test_data is None:
                # Splits the dataset to create training and test datasets
                if test_ratio is None:
                    test_ratio = 0.2
                if kwargs.get("sort_before_split", True):
                    (
                        X_train,
                        X_test,
                        Y_train,
                        Y_test,
                    ) = sort_train_test_split_shuffle(X, Y, test_ratio=test_ratio, seed=1)
                else:
                    X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size=test_ratio, random_state=1)
            else:
                # Uses the supplied test dataset
                X_train = X
                X_test = test_data[0]
                Y_train = Y
                Y_test = test_data[1]

            # Performs hyperopt for each predictor type passed into this function and appends each respective best
            # model and hyperparameters
            for predictor_class in self.input_predictor_type:
                predictor_instance = predictor_class(variables, self)
                model, model_params = predictor_instance.hyperparam_opt(
                    X_train, Y_train, test_ratio, predictor_params_dict=self.predictor_params_dict, **kwargs
                )
                model_list.append(model)
                model_params_list.append(model_params)

            # Determines the test set prediction accuracy using each optimised predictor type
            for m in model_list:
                Y_test_pred, _ = m.predict(X_test)
                dict = pm.all_performance_metrics(Y_test, Y_test_pred)
                test_loss_list.append(dict["RMSE"])

            # Identifies which predictor type had the most accurate test set prediction
            best_index = test_loss_list.index(min(test_loss_list))
            self.predictor_type = self.input_predictor_type[best_index]
            self.model_params = model_params_list[best_index]

        # The always_hyperparam_opt attribute is from the predictor/model. When True, the previously identified best
        # predictor type undergoes hyperparameter optimisation every Bayesian optimisation iteration
        elif self.obj_function.always_hyperparam_opt:
            # No model search; only hyperopt for the current best predictor type (identified previously)
            predictor_instance = self.predictor_type(variables, self)
            _, self.model_params = predictor_instance.hyperparam_opt(
                X, Y, test_ratio, predictor_params_dict=self.predictor_params_dict, **kwargs
            )

        # Re-fits the model but with the complete dataset
        # If test data was supplied, it is appended to the training data for the final model fitting
        if test_data is not None:
            X = np.vstack((X, X_test))
            Y = np.vstack((Y, Y_test))

        self.fit_regressor(X, Y, variables, test_ratio=test_ratio, **kwargs)

    def parity_plot(self) -> None:
        """

        Creates a parity plot for evaluating the fit quality of the regression model for the objective

        """
        # The include_validation attribute is from the predictor/model. When True, the X and Y data are split into
        # training and validation sets
        if self.obj_function.include_validation:
            # Produces a 2D scatter plot showing the parity plot data points, and the x = y line as a line plot
            train_paritydata = self.obj_function.performance_metrics["Train Parity Data"]
            train_rmse = self.obj_function.performance_metrics["Train RMSE"]
            train_r2 = self.obj_function.performance_metrics["Train r2"]
            val_paritydata = self.obj_function.performance_metrics["Validation Parity Data"]
            val_rmse = self.obj_function.performance_metrics["Validation RMSE"]
            val_r2 = self.obj_function.performance_metrics["Validation r2"]

            scatter_parity_data = [train_paritydata, val_paritydata]
            error = [
                self.obj_function.Y_train_error,
                self.obj_function.Y_val_error,
            ]
            max_parity = np.amax(np.vstack([train_paritydata, val_paritydata]))
            min_parity = np.amin(np.vstack([train_paritydata, val_paritydata]))
            # Array for the x = y line in the parity plot
            x_equals_y = [
                [min_parity, min_parity],
                [max_parity, max_parity],
            ]
            scatter_legend = (
                f"Train ({train_paritydata.shape[0]} experiments) (95% CI)",
                f"Validation ({val_paritydata.shape[0]} experiments) (95% CI)",
            )
            plot_title = (
                f"{self.name} Parity Plot ({self.obj_function.name} Model)\nTrain RMSE = "
                f"{round(train_rmse.astype('float'), 2)} {self.units}, Train r2 = {round(train_r2, 2)}, \n"
                f"Validation RMSE = {round(val_rmse.astype('float'), 2)} {self.units}, Validation r2 = "
                f"{round(val_r2, 2)}"
            )
            path = os.path.join(os.getcwd(), "ML Models", f"{self.obj_function.name}", "Parity Plots")
            if not os.path.exists(path):
                os.makedirs(path)
            output_file = os.path.join(
                path,
                rf"{self.name} Parity Plot ({self.obj_function.name} Model), "
                rf"{datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}",
            )

        else:
            train_paritydata = self.obj_function.performance_metrics["Parity Data"]
            train_rmse = self.obj_function.performance_metrics["RMSE"]
            train_r2 = self.obj_function.performance_metrics["r2"]
            scatter_parity_data = train_paritydata
            error = self.obj_function.Y_pred_error
            max_parity = np.amax(np.vstack([train_paritydata]))
            min_parity = np.amin(np.vstack([train_paritydata]))
            # Array for the x = y line in the parity plot
            x_equals_y = [
                [min_parity, min_parity],
                [max_parity, max_parity],
            ]
            scatter_legend = f"Train ({train_paritydata.shape[0]} experiments) (95% CI)"
            plot_title = (
                f"{self.name} Parity Plot ({self.obj_function.name} Model)\nRMSE = "
                f"{round(train_rmse.astype('float'), 2)} {self.units}, r2 = {round(train_r2, 2)}"
            )
            path = os.path.join(os.getcwd(), "ML Models", f"{self.obj_function.name}", "Parity Plots")
            if not os.path.exists(path):
                os.makedirs(path)
            output_file = os.path.join(
                path,
                rf"{self.name} Parity Plot ({self.obj_function.name} Model), "
                rf"{datetime.now().strftime('%d-%m-%Y, %H-%M-%S')}",
            )

        plotter.plot(
            plot_dim="2D",
            scatter_data=scatter_parity_data,
            error=error,
            line_data=x_equals_y,
            scatter_legend=scatter_legend,
            line_legend="x = y",
            xlabel=f"Actual {self.name} ({self.units})" if self.units != "" else f"Actual {self.name}",
            ylabel=f"Predicted {self.name} ({self.units})" if self.units != "" else f"Actual {self.name}",
            plottitle=plot_title,
            output_file=output_file,
        )
        plt.close()

    def partial_dependence_plot(self, X: np.ndarray) -> Dict[str, pd.DataFrame]:
        """

        Creates a partial dependence plot of the fitted model using the provided X array

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        Returns
        -------
        pdp_dict: Dict[str, pd.DataFrame]
            Dictionary that contains a DataFrame of the partial dependence plot data for every feature

        """
        partial_dependence_plot_dict = self.obj_function.partial_dependence_plot(X)

        return partial_dependence_plot_dict

    def cv(self, X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, **kwargs) -> Dict[str, Any]:
        """

        Function that is called to start the cross validation procedure using the fitted model hyperparameters and
        returns a dictionary containing the fitting results with emphasis on the test results. The number of k-folds is
        determined by the reciprocal of the test_ratio

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split into the validation and test sets where applicable

        Returns
        -------
        Dict[str, Any]
        Dictionary containing the models, their respective prediction quality for all k-folds and average cv statistics
        across all k-folds

        """
        X, Y = remove_nan(X, Y)

        cv_results = self.obj_function.cv(
            X,
            Y,
            self.model_params,
            test_ratio=test_ratio,
            **kwargs,
        )

        return cv_results

    def y_scrambling_cv(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float = 0.2,
        plot_parity: bool = True,
        inc_error_bars: bool = False,
        **kwargs,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """

        Compares the model fitting and accuracy of test data using cross validation for the supplied X and Y data set
        and a Y-scrambled array

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split into the validation and test sets where applicable
        plot_parity: bool, Default = True,
            Whether to plot a scatter plot of the Y test data points with predictions from the unshuffled and shuffled
            Y array
        inc_error_bars: bool, Default = False,
            Whether to include 95% confidence interval error bars on the scatter plot points

        Returns
        -------
        cv_results: Dict[str, Any]
            Contains the performance metrics from the fitted models using the original data set with emphasis on the
            test data results
        cv_results_y_scrambling: Dict[str, Any]
            Contains the performance metrics from the fitted models using the Y-scrambled array, with emphasis on
            the test data Results

        """
        X, Y = remove_nan(X, Y)

        cv_results, cv_results_y_scrambling = self.obj_function.y_scrambling_cv(
            X,
            Y,
            self.model_params,
            test_ratio=test_ratio,
            plot_parity=plot_parity,
            inc_error_bars=inc_error_bars,
            **kwargs,
        )

        return cv_results, cv_results_y_scrambling


class DeterministicFunction:
    """

    Inherit this DeterministicFunction class to write the calculable function

    Parameters
    ----------
    obj_function_data: Any
        Information that is required for the evaluate function

    """

    def __init__(self, obj_function_data: Optional[Any] = None):
        self.obj_function_data = obj_function_data
        self.name = "deterministic_function"

    def evaluate(self, X: np.ndarray, *args, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """

        The script to calculate and return the corresponding output values and standard deviations using the inputted
        X array

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        """
        pass


@define(kw_only=True)
class CalculableObjective(Objective):
    """

    Class for a calculable objective

    Parameters
    ----------
    name: str
        Name of the objective
    obj_max_bool: bool
        Whether the objective is to be maximised (True) or minimised (False)
    lower_bound: int | float
        The lower bound of the objective
    upper_bound: int | float
        The upper bound of the objective
    transformer: Optional. Default = None
        The transformation that should be applied to the objective for modelling. When left as the default
        None, the modelling uses the default transformations specified by the model type. Can be specified as the
        strings "normalisation", "standardisation", or "none" for min-max normalisation, standardisation to a mean of 0
        and a standard deviation of 1, or no transformation respectively. Alternatively, a class instance with
        fit_transform, transform, and inverse_transform functions, such as a scikit-learn transformer can be provided
        directly too.
    units: str, Default = ""
        The units for the objective
    obj_function: DeterministicFunction, Default = None
        Contains the information and script to calculate the output values and standard deviations of the calculable
        objective.

    """

    obj_function: Optional[DeterministicFunction] = None

    def calculate(self, X: np.ndarray, *args, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function to calculate the output values and standard deviations of the calculable objective

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        Returns
        -------
        Y: np.ndarray
            The calculated output values
        Y_stddev: np.ndarray
            The calculated standard deviations of the output values. Note: this is often equal to 0 for calculable
            objectives

        """
        Y, Y_stddev = self.obj_function.evaluate(X, *args, **kwargs)

        return Y.astype("float"), Y_stddev

    def draw_samples(self, X: np.ndarray, *args, **kwargs) -> np.ndarray:
        """

        Generates samples from a normal distribution when the X array passed

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        Returns
        -------
        np.ndarray
            Drawn samples using the model and the inputted X array. The outputted values have been un-transformed

        """
        Y_pred, Y_pred_stddev = self.calculate(X, *args, **kwargs)

        return np.random.default_rng().normal(Y_pred, Y_pred_stddev).astype("float")

    def add_obj_function(self, obj_function: DeterministicFunction) -> None:
        """

        If the extract_excel functionality was used to import a calculable objective, this function can be used to add
        the calculable objective information and script

        Parameters
        ----------
        obj_function: DeterministicFunction
            Contains the information and script to calculate the output values and standard deviations of the calculable
            objective

        """
        self.obj_function = obj_function


@define
class ObjectivesList:
    """

    Class that stores all objectives and their respective properties

    Parameters
    ----------
    objectives: List[Objective]
        A list of RegressionObjective and/or CalculableObjective objects

    """

    objectives: List[Objective]
    n_obj: int = field(init=False)
    names: List[str] = field(init=False)
    units: List[str] = field(init=False)
    max_bool_dict: Dict[str, bool] = field(init=False)
    bounds: np.ndarray = field(init=False)
    predictor_types: Dict[str, str] = field(init=False)

    def __attrs_post_init__(self):
        self.n_obj = len(self.objectives)
        self.names = [obj.name for obj in self.objectives]
        self.units = [obj.units for obj in self.objectives]
        self.max_bool_dict = {obj.name: obj.obj_max_bool for obj in self.objectives}
        self.bounds = np.array(
            [
                [obj.lower_bound for obj in self.objectives],
                [obj.upper_bound for obj in self.objectives],
            ]
        ).T
        self.predictor_types = {}

    def fit(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        model_search_bool: Optional[bool] = None,
        test_ratio: Optional[float] = None,
        test_data: Optional[List[np.ndarray]] = None,
        **kwargs,
    ) -> None:
        """

        Function that fits all regression objectives, including identifying suitable models and hyperparameters when
        applicable

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray
            Y array that contains the unprocessed objective data for all calculable and regression objectives, i.e.
            The data has not yet undergone any transformations related to standardisation/normalisation
        variables: VariablesList
            VariablesList object that contains information about all variables
        model_search_bool: bool
            Whether automated model and hyperparameter optimisation is to be performed during fitting the regression
            models
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split for the validation and test sets where applicable
        test_data: List[np.ndarray], Default = None,
            Test data to use for the initial model and hyperparameter optimisation, instead of creating it
            automatically from the provided X and Y arrays. The test X array is at index 0 and the test Y array is at
            index 1

        """
        if model_search_bool is None:
            model_search_bool = kwargs.get("model_search_bool", False)

        # Fit the ML regression models
        for obj_index, obj in enumerate(self.objectives):
            if isinstance(obj, RegressionObjective):
                obj.model_and_hyperparameter_opt(
                    X,
                    Y[:, obj_index],
                    variables,
                    model_search_bool=model_search_bool,
                    test_ratio=test_ratio,
                    test_data=test_data,
                    **kwargs,
                )
                self.predictor_types[self.names[obj_index]] = self.objectives[obj_index].obj_function.name

            else:
                self.predictor_types[self.names[obj_index]] = "calculable objective"

        logger.info(f"The identified best objective functions: {self.predictor_types}")

    def evaluate(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that uses the inputted X array to predict the RegressionObjective and/or calculate the
        CalculableObjective outputs

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        Returns
        -------
        Y: np.ndarray
            The output values obtained from the regression and/or calculable objectives
        Y_stddev: np.ndarray
            The standard deviations obtained from the regression and/or calculable objectives

        """
        Y = np.zeros((X.shape[0], len(self.objectives)), dtype=np.float64)
        Y_stddev = np.zeros_like(Y)
        for obj_index, obj in enumerate(self.objectives):
            if isinstance(obj, RegressionObjective):
                Y[:, obj_index], Y_stddev[:, obj_index] = obj.predict(X)
            elif isinstance(obj, CalculableObjective):
                Y[:, obj_index], Y_stddev[:, obj_index] = obj.calculate(X)

        return Y.astype("float"), Y_stddev.astype("float")

    def parity_plot(self) -> None:
        """

        Creates a parity plot for evaluating the fit quality of all regression models

        """
        for obj in self.objectives:
            if isinstance(obj, RegressionObjective):
                obj.parity_plot()

    def partial_dependence_plot(self, X: np.ndarray) -> Dict[str, Dict[str, pd.DataFrame]]:
        """

        Creates a partial dependence plots for all fitted regression models using the provided X array

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        Returns
        -------
        partial_dependence_plots_dict: Dict[str, Dict[str, pd.DataFrame]]
            Dictionary where the keys are the names of the objectives and the values are dictionaries that each contain
            a DataFrame of the partial dependence plot data for every feature

        """
        partial_dependence_plots_dict = {}
        for obj in self.objectives:
            if isinstance(obj, RegressionObjective):
                partial_dependence_plots_dict[obj.name] = obj.partial_dependence_plot(X)

        return partial_dependence_plots_dict

    def feature_importance(
        self, X: np.ndarray, Y: np.ndarray, transform_cat_to_cont: bool = True
    ) -> Dict[str, pd.Series]:
        """

        Returns a dictionary that shows the permutation feature importance for every machine learning model

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        transform_cat_to_cont: bool, Default = True
            If True, the permutation feature importance will be calculated for every descriptor of the
            CategoricalVariableWithDescriptors variables. When False, the permutation feature importance will be
            calculated for the categorical level directly

        Returns
        -------
        feature_importance_dict: Dict[str, pd.Series]
            Dictionary where the keys are the names of the objectives and the values are a pd.Series of the permutation
            feature importance for every feature

        """
        feature_importance_dict = {}
        for obj in self.objectives:
            if isinstance(obj, RegressionObjective):
                # Calculate permutation feature importance
                if isinstance(obj.predictor_type, NNBayesianModel) or isinstance(
                    obj.predictor_type, NNConcreteDropoutModel
                ):
                    repeats = 20
                else:
                    repeats = 50

                (obj.permutation_feature_importance, _,) = obj.obj_function.permutation_feature_importance(
                    X, Y, repeats=repeats, transform_cat_to_cont=transform_cat_to_cont
                )

                feature_importance_dict[obj.name] = obj.permutation_feature_importance

        return feature_importance_dict

    def cv(self, X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, **kwargs) -> Dict[str, Dict[str, Any]]:
        """

        Function that is called to start the cross validation procedure for every regression objective using their
        respective fitted model hyperparameters and returns a dictionary containing the fitting results with emphasis
        on the test results for each. The number of k-folds is determined by the reciprocal of the test_ratio

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split into the validation and test sets where applicable

        Returns
        -------
        cv_dict: Dict[str, Dict[str, Any]]
            Dictionary where the keys are the names of the objectives and the values are dictionaries containing the
            models, their respective prediction quality for all k-folds and average cv statistics across all k-folds

        """
        cv_dict = {}
        for obj_index, obj in enumerate(self.objectives):
            if isinstance(obj, RegressionObjective):
                cv_dict[obj.name] = obj.cv(X, Y[:, obj_index], test_ratio=test_ratio, **kwargs)

        return cv_dict

    def y_scrambling_cv(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float = 0.2,
        plot_parity: bool = True,
        inc_error_bars: bool = False,
        **kwargs,
    ) -> Tuple[Dict[str, Dict[str, Any]], Dict[str, Dict[str, Any]]]:
        """

        Compares the model fitting and accuracy of test data using cross validation for the supplied X and Y data set
        and a Y-scrambled array of all regression objectives

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float, Default = 0.2,
            The proportion of inputted X and Y arrays to be split into the validation and test sets where applicable
        plot_parity: bool, Default = True,
            Whether to plot a scatter plot of the Y test data points with predictions from the unshuffled and shuffled
            Y array
        inc_error_bars: bool, Default = False,
            Whether to include 95% confidence interval error bars on the scatter plot points

        Returns
        -------
        cv_dict: Dict[str, Any]
            Dictionary where the keys are the names of the objectives and the values contain the performance metrics
            from the fitted models using the original data set with emphasis on the test data results
        cv_y_scrambling_dict: Dict[str, Any]
            Dictionary where the keys are the names of the objectives and the values contain the performance metrics
            from the fitted models using the Y-scrambled array, with emphasis on the test data Results

        """
        cv_dict, cv_y_scrambling_dict = {}, {}
        for obj_index, obj in enumerate(self.objectives):
            if isinstance(obj, RegressionObjective):
                cv_dict[obj.name], cv_y_scrambling_dict[obj.name] = obj.y_scrambling_cv(
                    X,
                    Y[:, obj_index],
                    test_ratio=test_ratio,
                    plot_parity=plot_parity,
                    inc_error_bars=inc_error_bars,
                    **kwargs,
                )

        return cv_dict, cv_y_scrambling_dict

    def append(self, obj: Objective):
        """

        Function for appending an additional regression or calculable objective to the current ObjectiveList object

        Parameters
        ----------
        obj: Objective
            Regression or calculable objective

        """
        obj_list = copy.copy(self.objectives)
        obj_list.append(obj)

        return ObjectivesList(obj_list)

    def add_calculable_obj_function(self, name: str, obj_function: DeterministicFunction) -> None:
        """

        If the extract_excel functionality was used to import a calculable objective, this function can be used to add
        the calculable objective information and script to the named calculable objective

        Parameters
        ----------
        name: str
            Name of the calculable objective to add the obj_function data to
        obj_function: DeterministicFunction
            Contains the information and script to calculate the output values and standard deviations of the
            calculable objective

        """
        for obj in self.objectives:
            if isinstance(obj, CalculableObjective):
                if name == obj.name:
                    obj.obj_function = obj_function
                    break

    def draw_samples(self, X: np.ndarray) -> np.ndarray:
        """

        Function that uses the inputted X array to draw samples from the distributions of the RegressionObjective
        and/or CalculableObjective outputs

        Parameters
        ----------
        X: np.ndarray
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation

        Returns
        -------
        Y: np.ndarray
            The output values obtained from the regression and/or calculable objectives

        """
        Y = np.zeros((X.shape[0], len(self.objectives)), dtype=np.float64)
        for obj_index, obj in enumerate(self.objectives):
            Y[:, obj_index] = obj.draw_samples(X)

        return Y.astype("float")



import os
from dataclasses import dataclass
from typing import Optional, Tuple

import numpy as np
from scipy import spatial

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList
from nemo_bo.opt.samplers import LatinHyperCubeSampling, PoolBased, SampleGenerator
from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class HighestUncertainty(AcquisitionFunction):
    """

    For single-objective and multi-objective problems, this acquisition function is an explorative method that
    identifies the approximate region of the variable space that has the highest uncertainty in the objective
    prediction

    Parameters
    ----------
    num_candidates: int
        The number of sets of X arrays to be suggested by the acquisition function

    """

    num_candidates: int

    def generate_candidates(
        self,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: SampleGenerator,
        constraints: Optional[ConstraintsList] = None,
        num_new_points: int = 2**21,
        **acq_func_kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that is called to generate the candidates that has Y-values with the highest uncertainty

        Parameters
        ----------
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        sampler: SampleGenerator
            Used to generate samples (SampleGenerator). PoolBased is not supported
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints
        num_new_points: int, Default: 2 ** 21
            Number of new sets of input variables to generate

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest uncertainty
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest uncertainty

        """
        if isinstance(sampler, PoolBased):
            raise NotImplementedError(
                "The PoolBased sampler is incompatible with the HighestUncertainty acquisition function. Please select "
                "a different sampler if you wish to continue using this acquisition function. Alternatively, you may "
                "use the ExpectedImprovement acquisition function with the PoolBased sampler instead"
            )

        logger.info(f"Identifying conditions with the highest uncertainty")

        if sampler is None:
            sampler = LatinHyperCubeSampling(num_new_points)
        else:
            sampler.num_new_points = num_new_points

        X_new_points = sampler.generate_samples(variables, constraints)

        Y_new_points, Y_new_points_stddev = objectives.evaluate(X_new_points)

        # Min_max normalisation of Y standard deviation
        # I think this should be fine to minmax normalise like this
        for obj_index, _ in enumerate(objectives.bounds):
            Y_new_points_stddev[:, obj_index] = Y_new_points_stddev[:, obj_index] / (
                objectives.bounds[obj_index, 1] - objectives.bounds[obj_index, 0]
            )

        Y_new_points_stddev_sum = np.sum(Y_new_points_stddev, axis=1).reshape(-1, 1)
        sorting_indices = Y_new_points_stddev_sum[:, -1].argsort()

        X_new_points_sorted = X_new_points[sorting_indices][::-1]

        if variables.num_cat_var > 0:
            X_new_points = variables.categorical_transform(X_new_points).astype("float")
        X_new_points_norm_sorted = (
            (X_new_points[sorting_indices][::-1] - np.array(variables.lower_bounds))
            / (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
        )[sorting_indices][::-1]
        Y_new_points_sorted = Y_new_points[sorting_indices][::-1]
        Y_new_points_stddev_sum_sorted = Y_new_points_stddev_sum[sorting_indices][::-1]

        # 0 index for the highest uncertainty one at the top of the array
        chosen_indices = [0]
        # cosine similarity index of 1.0 for the highest uncertainty one at the top of the array
        cosine_simularity_list = [1.0]
        # Comparing the X-values for the highest uncertainty with every other entry down the sorted X array
        for index, x in enumerate(X_new_points_norm_sorted):
            cosine_simularity_index = 1 - spatial.distance.cosine(X_new_points_norm_sorted[0], x)
            # Only passes if the cosine similarity index of a given x row is greater than 0.1 different compared to all
            # of the x's in the list
            if all((((cs - cosine_simularity_index) ** 2) ** 0.5) > 0.1 for cs in cosine_simularity_list):
                cosine_simularity_list.append(cosine_simularity_index)
                chosen_indices.append(index)

            if len(chosen_indices) == self.num_candidates:
                break

        selected_X = X_new_points_sorted[chosen_indices]
        selected_Y = Y_new_points_sorted[chosen_indices].astype(np.float64)
        total_Y_std = Y_new_points_stddev_sum_sorted[chosen_indices].astype(np.float64)

        logger.info(f"Identified variables with the most uncertain output values")

        return selected_X, selected_Y



import os
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList
from nemo_bo.opt.samplers import LatinHyperCubeSampling, PoolBased, SampleGenerator
from nemo_bo.opt.variables import VariablesList
import nemo_bo.utils.logger as logging_nemo

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class AcquisitionFunction:
    """

    Base class for acquisition functions that should be inherited from

    """

    def generate_candidates(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: Union[SampleGenerator, PoolBased],
        constraints: Optional[ConstraintsList] = None,
        *args,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that is called to generate the candidates using the acquisition function

        """
        pass

    def build_ref_point(self, max_bool_dict: Dict[str, bool]) -> List[float]:
        """

        Function that creates the normalised reference point for hypervolumne calculations

        Parameters
        ----------
        max_bool_dict: Dict[str, bool]
            A dictionary where the keys are the objective names and the values are whether the objective is to be
            maximised (True) or minimised (False)

        """
        ref_point = []
        hv_obj_minmax = np.array([[-0.01, 1.01] for _ in enumerate(max_bool_dict)])
        for obj_index, obj in enumerate(max_bool_dict):
            if max_bool_dict[obj]:
                ref_point.append(hv_obj_minmax[obj_index][0])
            else:
                ref_point.append(hv_obj_minmax[obj_index][1] * -1)

        return ref_point

    def Y_norm_minmax_transform(self, Y: np.ndarray, bounds: np.ndarray, max_bool_dict: Dict[str, bool]) -> np.ndarray:
        """

        Function that min-max normalisation transforms the objective values and multiplies by -1 for minimising
        objectives

        Parameters
        ----------
        Y: np.ndarray
            Y array to be transformed
        bounds: np.ndarray

        max_bool_dict: Dict[str, bool]
            A dictionary where the keys are the objective names and the values are whether the objective is to be
            maximised (True) or minimised (False)

        Returns
        -------
        sign_adjusted_Y: np.ndarray
            Transformed Y array

        """
        # Transforms Y using min-max normalisation
        sign_adjusted_Y = np.zeros_like(Y)
        for obj_index, obj in enumerate(max_bool_dict):
            if Y.ndim > 1:
                sign_adjusted_Y[:, obj_index] = (Y[:, obj_index] - bounds[obj_index, 0]) / (
                    bounds[obj_index, 1] - bounds[obj_index, 0]
                )

                #
                # Maximising objective values are left unchanged and minimising objective values are multiplied by -1
                if not max_bool_dict[obj]:
                    sign_adjusted_Y[:, obj_index] *= -1
            else:
                sign_adjusted_Y[obj_index] = (Y[obj_index] - bounds[obj_index, 0]) / (
                    bounds[obj_index, 1] - bounds[obj_index, 0]
                )

                # maximising objective values are left unchanged and minimising objective values are multiplied by -1
                if not max_bool_dict[obj]:
                    sign_adjusted_Y[obj_index] *= -1

        return sign_adjusted_Y



import sys
from typing import Any, Dict

from keras.models import Model


class SaveModelCheckPoint:
    """

    Class that enables saving of tensorflow neural network models based on both the train loss and validation loss
    values

    Parameters
    ----------
    model : KerasRegressor type object that contains the neural network being trained
    filepath : str type of the file path to save the .h5 file that stores the model parameters

    """

    def __init__(self, model: Model, filepath: str):
        self.model = model
        self.filepath = filepath
        self.reset()

    def reset(self) -> None:
        """

        Function to reset the values that are monitored during the checkpointing process. This function can be called
        at the end of model training to allow multiple models to be trained using the same class instance of
        SaveModelCheckPoint. An example use-case would be during cross validation

        """
        self.best_epoch = 0
        self.best_loss = sys.float_info.max
        self.best_val_loss = sys.float_info.max

    def save_weights(self, epoch: int, logs: Dict[str, Any]) -> None:
        """

        Function called during the model training procedure when a LambdaCallback is used that allows loss monitoring
        and model savings

        Parameters
        ----------
        epoch: int
            The current training epoch
        logs: Dict[str, Any]
            Contains loss information for the current training epoch

        """
        loss = logs["loss"]
        val_loss = logs["val_loss"]

        if epoch == 0:
            for _ in range(10):
                try:
                    self.model.save_weights(self.filepath)
                    break
                except OSError as e:
                    print(e)

        elif loss < self.best_loss and val_loss < self.best_val_loss:
            self.best_loss = loss
            self.best_val_loss = val_loss
            self.best_epoch = epoch
            # print(
            #     f"This epoch's loss ({loss}) and validation loss ({val_loss}) are both the best so far. Saving the "
            #     f"model weights.\n"
            # )
            for _ in range(10):
                try:
                    self.model.save_weights(self.filepath)
                    break
                except OSError as e:
                    print(e)
        # else:
        #     print(f"Number of epochs since the weights were last saved = {epoch - self.best_epoch}.\n")



# Utilises the BoToroch package for creating GP models
# M. Balandat, B. Karrer, D. R. Jiang, S. Daulton, B. Letham, A. G. Wilson, and E. Bakshy. BoTorch: A Framework for
# Efficient Monte-Carlo Bayesian Optimization. Advances in Neural Information Processing Systems 33, 2020.
# http://arxiv.org/abs/1910.06403
# https://github.com/pytorch/botorch
# https://botorch.org/

from __future__ import annotations

import copy
import os
import time
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import torch
from botorch.models import SingleTaskGP
from botorch.fit import fit_gpytorch_model
from gpytorch.mlls import ExactMarginalLogLikelihood
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import AdditiveStructureKernel, Kernel, MaternKernel, ScaleKernel
from gpytorch.likelihoods import Likelihood
from gpytorch.means import ConstantMean
from gpytorch.models import ExactGP
from gpytorch.priors.torch_priors import GammaPrior
from torch import Tensor

import nemo_bo.utils.logger as logging_nemo
import nemo_bo.utils.perf_metrics as pm
from nemo_bo.models.base.base_model import Base_Model

if TYPE_CHECKING:
    from nemo_bo.opt.objectives import RegressionObjective
    from nemo_bo.opt.variables import VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


class ExactGPModel(ExactGP):
    """

    Defines the GP model and forward function for regression. Details about the parameters can be found within the
    GPyTorch package documentation

    """

    def __init__(self, trainX: Tensor, trainY: Tensor, likelihood: Likelihood, kernel: Kernel):
        super(ExactGPModel, self).__init__(trainX, trainY, likelihood)
        self.mean_module = ConstantMean()
        self.covar_module = kernel

    def forward(self, x: Tensor) -> MultivariateNormal:
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


class GPModel(Base_Model):
    """

    Used to create a GP model for the RegressionObjective

    See docstring for Base_Model __init__ function for information

    """

    def __init__(
        self, variables: VariablesList, objective: RegressionObjective, always_hyperparam_opt: bool = True, **kwargs
    ):
        super().__init__(variables, objective, always_hyperparam_opt)
        self.default_X_transform_type = "normalisation"
        self.default_Y_transform_type = "standardisation"
        self.include_validation = False
        self.name = "gp"

    def fit(self, X: np.ndarray, Y: np.ndarray, **params) -> None:
        """

        Function that is called to start the GP model fitting procedure

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        params: Dict[str, Any]
            Dictionary that contains the hyperparameters for the model

        """
        if self.variables.num_cat_var > 0:
            self.X_train = self.variables.categorical_transform(X).astype("float")
        else:
            self.X_train = X

        self.X_train, self.Y_train = self.transform_by_predictor_type(self.X_train, Y=Y)

        self.fit_model(params)

        self.Y_pred, self.Y_pred_stddev = self.predict(X)
        self.performance_metrics = pm.all_performance_metrics(Y, self.Y_pred)

        self.Y_pred_error = 1.96 * self.Y_pred_stddev

    def fit_model(self, params: Dict[str, Any]) -> None:
        """

        Called by the fit and cv-related functions for fitting the GP model using pre-processed X and Y training data
        in the class instance

        """
        X_tensor = torch.tensor(
            self.X_train,
            device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
        )
        Y_tensor = torch.tensor(
            self.Y_train,
            device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
        )
        if Y_tensor.ndim == 1:
            Y_tensor = Y_tensor.reshape(-1, 1)

        self.n_iterations = int(params.get("n_iterations", 50))
        self.kernel = params.get(
            "kernel",
            ScaleKernel(
                MaternKernel(nu=2.5, ard_num_dims=X_tensor.shape[-1], lengthscale_prior=GammaPrior(3.0, 6.0)),
                outputscale_prior=GammaPrior(2.0, 0.15),
            ),
        )

        self.model = SingleTaskGP(X_tensor, Y_tensor, covar_module=copy.deepcopy(self.kernel))
        if torch.cuda.is_available():
            self.model = self.model.to("cuda")

        mll = ExactMarginalLogLikelihood(self.model.likelihood, self.model)
        for _ in range(5):
            try:
                fit_gpytorch_model(mll)
            except AttributeError as e:
                print(e)

    def predict(self, X: np.ndarray, X_transform: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """

        Determines the model output and its standard deviation of the X array passed into the function

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation
        X_transform: bool, Default = True
            For the default models provided natively in NEMO, the initial transformation of categorical variables and
            transformations related to standardisation/normalisation can be switched off by passing X_transform = False

        Returns
        ----------
        Y_pred: np.ndarray
            The predictive mean of the objective calculated using the model and the inputted X array. The outputted
            values have been un-transformed
        Y_pred_stddev: np.ndarray
            The predictive standard deviation of the objective calculated using the model and the inputted X array. The
            outputted values have been un-transformed

        """
        # X needs to be a 2D array
        if X.ndim == 1:
            X = X.reshape(1, -1)

        if X_transform:
            if self.variables.num_cat_var > 0:
                X = self.variables.categorical_transform(X).astype("float")
            X = self.transform_only_by_predictor_type(X)

        # Split to reduce memory requirements. I think this is not ideal for the quality of the prediction but it helps
        # to prevent crashing caused by memory limitations when the X array is too large
        split_size = 32000
        X_split = np.split(X, np.array((range(1, int(X.shape[0] / split_size) + 1))) * split_size)
        observed_Y_pred_list = []
        for X_split_each in X_split:
            if X_split_each.shape[0] == 0:
                continue
            with torch.no_grad():
                observed_Y_pred = self.model.posterior(
                    torch.tensor(
                        X_split_each,
                        device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
                    )
                )

            observed_Y_pred_list.append(observed_Y_pred)

        Y_pred = np.vstack([observed_Y_pred.mean.cpu().detach().numpy() for observed_Y_pred in observed_Y_pred_list])
        Y_pred_upper = np.hstack(
            [
                observed_Y_pred.mvn.confidence_region()[1].cpu().detach().numpy()
                for observed_Y_pred in observed_Y_pred_list
            ]
        )
        Y_pred_upper = Y_pred_upper.reshape(Y_pred.shape)

        Y_pred = self.objective.inverse_transform(Y_pred)
        Y_pred_upper = self.objective.inverse_transform(Y_pred_upper)
        Y_pred_stddev = (np.subtract(Y_pred_upper, Y_pred)) / 1.96

        return Y_pred.flatten(), Y_pred_stddev.flatten()

    def draw_samples(self, X: np.ndarray, X_transform: bool = True) -> np.ndarray:
        """

        Generates samples from the distribution using the X array

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be used in the prediction, i.e. The data has not
            yet undergone any transformations related to converted categorical variables to their respective
            descriptors or transformations related to standardisation/normalisation
        X_transform: bool, Default = True
            For the default models provided natively in NEMO, the initial transformation of categorical variables and
            transformations related to standardisation/normalisation can be switched off by passing X_transform = False

        Returns
        ----------
        Y_samples: np.ndarray
            Drawn samples of the objective using the model and the inputted X array. The outputted values have been
            un-transformed

        """
        # X needs to be a 2D array
        if X.ndim == 1:
            X = X.reshape(1, -1)

        if X_transform:
            if self.variables.num_cat_var > 0:
                X = self.variables.categorical_transform(X).astype("float")
            X = self.transform_only_by_predictor_type(X)

        # Split to reduce memory requirements. I think this is not ideal for the quality of the prediction but it helps
        # to prevent crashing caused by memory limitations when the X array is too large
        split_size = 50000
        X_split = np.split(X, np.array((range(1, int(X.shape[0] / split_size) + 1))) * split_size)
        observed_Y_pred_list = []
        for X_split_each in X_split:
            if X_split_each.shape[0] == 0:
                continue
            with torch.no_grad():
                observed_Y_pred = self.model.posterior(
                    torch.tensor(
                        X_split_each,
                        device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
                    )
                )
            observed_Y_pred_list.append(observed_Y_pred)
        Y_pred = np.vstack(
            [observed_Y_pred.sample().cpu().detach().numpy()[0] for observed_Y_pred in observed_Y_pred_list]
        )
        Y_samples = self.objective.inverse_transform(Y_pred)

        return Y_samples.flatten()

    @staticmethod
    def default_params(n_var: int, gp_kernel_choices: Optional[List[Kernel]] = None) -> Dict[str, Any]:
        """

        Function that returns the default kernels to use for the optimisation of GP models

        Parameters
        ----------
        n_var: int
            Number of continuous variables
        gp_kernel_choices: List[Kernel], Default = None
            The types of kernels to use for the GP optimisation

        """
        if gp_kernel_choices is None:
            gp_kernel_choices = [
                AdditiveStructureKernel(
                    base_kernel=ScaleKernel(
                        MaternKernel(ard_num_dims=n_var, nu=0.5, lengthscale_prior=GammaPrior(3.0, 6.0)),
                        outputscale_prior=GammaPrior(2.0, 0.15),
                    ),
                    num_dims=n_var,
                ),
                AdditiveStructureKernel(
                    base_kernel=ScaleKernel(
                        MaternKernel(ard_num_dims=n_var, nu=1.5, lengthscale_prior=GammaPrior(3.0, 6.0)),
                        outputscale_prior=GammaPrior(2.0, 0.15),
                    ),
                    num_dims=n_var,
                ),
                AdditiveStructureKernel(
                    base_kernel=ScaleKernel(
                        MaternKernel(ard_num_dims=n_var, nu=2.5, lengthscale_prior=GammaPrior(3.0, 6.0)),
                        outputscale_prior=GammaPrior(2.0, 0.15),
                    ),
                    num_dims=n_var,
                ),
                ScaleKernel(
                    MaternKernel(nu=0.5, ard_num_dims=n_var, lengthscale_prior=GammaPrior(3.0, 6.0)),
                    outputscale_prior=GammaPrior(2.0, 0.15),
                ),
                ScaleKernel(
                    MaternKernel(nu=1.5, ard_num_dims=n_var, lengthscale_prior=GammaPrior(3.0, 6.0)),
                    outputscale_prior=GammaPrior(2.0, 0.15),
                ),
                ScaleKernel(
                    MaternKernel(nu=2.5, ard_num_dims=n_var, lengthscale_prior=GammaPrior(3.0, 6.0)),
                    outputscale_prior=GammaPrior(2.0, 0.15),
                ),
            ]

        return {
            "kernel": gp_kernel_choices,
        }

    def hyperparam_opt(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        test_ratio: float,
        predictor_params_dict: Optional[Dict[str, Dict[str, Callable]]] = None,
        gp_kernel_choices: Optional[list] = None,
        **kwargs,
    ) -> Tuple[GPModel, Dict[str, Any]]:
        """

        Function to be called for optimisation of GP models. Returns the model fitted with the best kernel and a
        dictionary containing the best kernel

        Parameters
        ----------
        X: np.ndarray,
            X array that contains the unprocessed variables data to be fitted to the model, i.e. The data has not yet
            undergone any transformations related to converted categorical variables to their respective descriptors
            or transformations related to standardisation/normalisation
        Y: np.ndarray,
            Y array that contains the unprocessed objective data, i.e. The data has not yet undergone any
            transformations related to standardisation/normalisation
        test_ratio: float
            The proportion of inputted X and Y arrays to be split for the validation set where applicable
        predictor_params_dict: Dict[str, Dict[str, Callable]], Default = None
            Dictionary with a key 'gp' and a corresponding value that is a dictionary that is used for GP optimisation
        gp_kernel_choices: List[Kernel], Default = None
            The types of kernels to use for the GP optimisation

        Returns
        -------
        model: GPModel
            The model fitted with the best kernel found

        model_params: Dict[str, Any]
            Dictionary containing the best kernel found

        """
        if predictor_params_dict is None:
            predictor_params_dict = self.default_params(self.variables.n_var, gp_kernel_choices)
        else:
            predictor_params_dict = predictor_params_dict.get(
                self.name, self.default_params(self.variables.n_var, gp_kernel_choices)
            )

        test_rmse_list = []
        kernel_list = []
        best_loss = None
        for index, kernel in enumerate(predictor_params_dict["kernel"]):
            start_time = time.perf_counter()

            cv_results = self.cv(X, Y, {"kernel": kernel}, test_ratio=test_ratio, **kwargs)

            test_rmse_list.append(cv_results["Mean Test RMSE"])
            kernel_list.append(kernel)

            if best_loss is None:
                best_loss = cv_results["Mean Test RMSE"]
            else:
                if cv_results["Mean Test RMSE"] < best_loss:
                    best_loss = cv_results["Mean Test RMSE"]

            print(
                f'{index + 1}/{len(predictor_params_dict["kernel"])}, iteration time (s) = '
                f'{time.perf_counter() - start_time:>4.2f}, current loss: {cv_results["Mean Test RMSE"]}, best loss: '
                f"{best_loss}"
            )

        best_kernel_index = test_rmse_list.index(min(test_rmse_list))
        model_params = {"kernel": kernel_list[best_kernel_index]}

        model = self.new_instance(self.__class__, **kwargs)
        model.fit(X, Y, test_ratio=test_ratio, **model_params, **kwargs)

        logger.info(f"Completed GP optimisation")
        return model, model_params



import os
from dataclasses import dataclass
from itertools import combinations
from typing import Optional, Tuple

import numpy as np
import torch
from botorch.utils.multi_objective.hypervolume import Hypervolume
from botorch.utils.multi_objective.pareto import is_non_dominated
from pymoo.core.result import Result
from pymoo.factory import get_reference_directions, get_termination
from pymoo.optimize import minimize

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.acquisition_functions.nsga_improvement.problem_wrapper import NSGAProblemWrapper
from nemo_bo.acquisition_functions.nsga_improvement.unsga3 import UNSGA3
from nemo_bo.acquisition_functions.highest_uncertainty.highest_uncertainty import HighestUncertainty
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.objectives import ObjectivesList
from nemo_bo.opt.samplers import LatinHyperCubeSampling, PoolBased, SampleGenerator
from nemo_bo.opt.variables import VariablesList
from nemo_bo.utils.data_proc import remove_all_nan_rows

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class NSGAImprovement(AcquisitionFunction):
    """

    For single-objective and multi-objective problems, this acquisition function can select between a method that
    identifies the highest hypervolume improvement using a U-NSGA-III algorithm for selecting candidates, or an
    explorative method that identifies the approximate region of the variable space that has the highest uncertainty
    in the objective prediction

    Parameters
    ----------
    num_candidates: int
        The number of sets of X arrays to be suggested by the acquisition function

    """

    num_candidates: int

    def generate_candidates(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: Optional[SampleGenerator] = None,
        constraints: Optional[ConstraintsList] = None,
        **acq_func_kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that is called when generate the best candidates using the U-NSGA-III algorithm or the highest
        uncertainty acquisition function

        Parameters
        ----------
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        sampler: SampleGenerator, Default = None
            Used to generate samples (SampleGenerator). PoolBased is not supported
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest hypervolume improvement or highest
            uncertainty
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest hypervolume improvement or highest
            uncertainty

        """
        if isinstance(sampler, PoolBased):
            raise NotImplementedError(
                "The PoolBased sampler is incompatible with the NSGAImprovement acquisition function. Please select a "
                "different sampler if you wish to continue using this acquisition function. Alternatively, you may use "
                "the ExpectedImprovement acquisition function with the PoolBased sampler instead"
            )

        if self.num_candidates > 4:
            logger.warning(
                f"The number of suggested candidates is more than 4, which is not recommended due to the risk of long "
                f"run times"
            )

        self.pop_size = acq_func_kwargs.get("pop_size")
        self.generations = acq_func_kwargs.get("generations")

        Y = remove_all_nan_rows(Y)

        if self.pop_size is None:
            self.pop_size = 200

        if self.generations is None:
            self.generations = 50

        self.generate_pareto(variables, objectives, constraints)
        selected_X, selected_Y = self.hvi(self.res.X, self.res.F, Y, variables, objectives)

        if not self.hvi_search_pass:
            highest_uncertainty = HighestUncertainty(num_candidates=self.num_candidates)
            selected_X, selected_Y = highest_uncertainty.generate_candidates(
                variables, objectives, sampler=sampler, constraints=constraints
            )

        return selected_X, selected_Y

    def generate_pareto(
        self, variables: VariablesList, objectives: ObjectivesList, constraints: ConstraintsList
    ) -> Result:
        """
        Runs the U-NSGA-III method to generate the samples that have the highest hypervolume

        Parameters
        ----------
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        res: Result
            Contains all of the information and results from the U-NSGA-III method

        """

        xl = 0
        xu = 1

        ref_dirs = get_reference_directions("energy", objectives.n_obj, self.pop_size * 0.9, seed=1)

        logger.info(f"Running the U-NSGA-III algorithm")
        algorithm = UNSGA3(ref_dirs, pop_size=self.pop_size)
        problem = NSGAProblemWrapper(xl, xu, variables, objectives, constraints)
        termination = get_termination("n_gen", self.generations)

        # Runs the U-NSGA-III algorithm
        self.res = minimize(problem, algorithm, termination, verbose=True)

        # Correct by * -1 for maximising objectives from the U-NSGA-III algorithm
        for obj_index, obj in enumerate(objectives.max_bool_dict):
            if objectives.max_bool_dict[obj]:
                if objectives.n_obj > 1:
                    self.res.F[:, obj_index] *= -1
                else:
                    self.res.F[obj_index] *= -1

        # Undo min-max normalisation of X values identified by U-NSGA-III
        self.res.X = (self.res.X * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))) + np.array(
            variables.lower_bounds
        )

        return self.res

    def hvi(
        self,
        X_nsga: np.ndarray,
        Y_nsga: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Identifies the candidate(s) from the solutions from the U-NSGA-III optimisation that will produce the highest
        hypervolume improvement

        Parameters
        ----------
        X_nsga: np.ndarray
            Variable values for the solutions from the U-NSGA-III optimisation
        Y_nsga : np.ndarray
            Objective values for the solutions from the U-NSGA-III optimisation
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest hypervolume improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest hypervolume improvement

        """
        logger.info(f"Identifying conditions with the largest hypervolume")

        # Construct hypervoume reference point
        ref_point = self.build_ref_point(objectives.max_bool_dict)

        # Min-max normalisation of objective values and * -1 for minimising objectives
        sign_adjusted_Y = self.Y_norm_minmax_transform(Y, objectives.bounds, objectives.max_bool_dict)
        sign_adjusted_Y_nsga = self.Y_norm_minmax_transform(Y_nsga, objectives.bounds, objectives.max_bool_dict)

        # Appends all Y_nsga points to the Y array
        Yappy_Ynsga = np.vstack([sign_adjusted_Y, sign_adjusted_Y_nsga])

        # Identifies which data points are non-dominated
        pareto_mask_Yappy_Ynsga = is_non_dominated(torch.tensor(Yappy_Ynsga, dtype=torch.double))

        # Slice out the mask of the Y_nsga points on the pareto front
        pareto_front_Ynsga_mask = pareto_mask_Yappy_Ynsga[Y.shape[0] :].cpu().detach().numpy()

        # If the number of Y_nsga points is greater than or equal to the desired number of candidates
        if pareto_front_Ynsga_mask.sum() >= self.num_candidates:
            # The X_nsga, Y_nsga and sign_adjusted_Y_nsga points that are on the pareto front
            X_nsga = X_nsga[pareto_front_Ynsga_mask]
            Y_nsga = Y_nsga[pareto_front_Ynsga_mask]
            sign_adjusted_Y_nsga = sign_adjusted_Y_nsga[pareto_front_Ynsga_mask]

            # Creates a list of hypervolumes for candidate(s) that successfully improve the hypervolume
            hv_list = []
            candidates_X_list = []
            candidates_Y_list = []
            X_nsga_combos = list(combinations(X_nsga, self.num_candidates))
            Y_nsga_combos = list(combinations(Y_nsga, self.num_candidates))
            sign_adjusted_Y_nsga_combos = list(combinations(sign_adjusted_Y_nsga, self.num_candidates))
            for x, y, candidates_Y in zip(X_nsga_combos, Y_nsga_combos, sign_adjusted_Y_nsga_combos):
                # Adds the candidate objective values to the experimental data set
                YappF = np.vstack([sign_adjusted_Y, np.array(candidates_Y)])
                # Identifies which data points are non-dominated
                pareto_mask_YappF = is_non_dominated(torch.tensor(YappF, dtype=torch.double))
                # Creates an numpy.ndarray containing the non-dominated points
                pareto_front_YappF = YappF[pareto_mask_YappF]

                candidates_X_list.append(x)
                candidates_Y_list.append(y)

                # Calculates the hypervolume
                hv = Hypervolume(ref_point=torch.tensor(ref_point, dtype=torch.double))
                hv_list.append(hv.compute(torch.tensor(pareto_front_YappF, dtype=torch.double)))

            # Selects the variable and objective values of the candidate(s) that produced the largest hypervolume
            max_idx = hv_list.index(max(hv_list))
            selected_X = np.array(candidates_X_list[max_idx])
            selected_Y = np.array(candidates_Y_list[max_idx])

            if variables.num_cat_descriptor_var > 0:
                selected_X = variables.descriptor_to_name(selected_X)

            self.hvi_search_pass = True
            logger.info("Conditions were successfully selected using hypervolume improvement")

        else:
            logger.info(
                "Failed to identify a set of conditions that would improve the hypervolume at the desired batch size"
            )
            selected_X = np.array([0])  # Dummy array
            selected_Y = np.array([0])  # Dummy array
            self.hvi_search_pass = False

        return selected_X, selected_Y
